"""# Image widget

Example of creating an Image Widget from a file.

(This requires pillow, or that magicgui was installed as ``magicgui[image]``)
"""

from magicgui.widgets import Image

image = Image(value="../../images/_test.jpg")
image.scale_widget_to_image_size()
image.show(run=True)



"""# Log slider widget

A logarithmic scale range slider widget.
"""

from magicgui import magicgui


@magicgui(
    auto_call=True,
    result_widget=True,
    input={"widget_type": "LogSlider", "max": 10000, "min": 1, "tracking": False},
)
def slider(input=1):
    """Logarithmic scale slider."""
    return round(input, 4)


slider.show(run=True)



"""Protocols (interfaces) for backends to implement.

Each class in this module is a [typing.Protocol][typing.Protocol] that specifies
a set of :func:`~abc.abstractmethod` that a backend widget must implement
to be compatible with the magicgui API.
All magicgui-specific abstract methods are prefaced with ``_mgui_*``.

For an example backend implementation, see ``magicgui.backends._qtpy.widgets``
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    NoReturn,
    Protocol,
    Sequence,
    runtime_checkable,
)

if TYPE_CHECKING:
    import numpy as np

    from magicgui.widgets.bases import Widget


def assert_protocol(widget_class: type, protocol: type) -> None:
    """Ensure that widget_class implements protocol, or raise helpful error."""
    if not isinstance(widget_class, protocol):
        _raise_protocol_error(widget_class, protocol)
    return None


def _raise_protocol_error(widget_class: type, protocol: type) -> NoReturn:
    """Raise a more helpful error when required protocol members are missing."""
    missing = {
        i
        for i in set(dir(protocol)) - set(dir(widget_class))
        if not i.startswith(("__", "_is_protocol", "_is_runtime", "_abc_impl"))
    }
    message = (
        f"{widget_class!r} does not implement {protocol.__name__!r}.\n"
        f"Missing methods: {missing!r}"
    )
    raise TypeError(message)


@runtime_checkable
class WidgetProtocol(Protocol):
    """Base Widget Protocol: specifies methods that all widgets must provide."""

    def __init__(self, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def _mgui_close_widget(self) -> None:
        """Close widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_visible(self) -> bool:
        """Get widget visibility."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_visible(self, value: bool) -> None:
        """Set widget visibility."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_enabled(self) -> bool:
        """Get the enabled state of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_enabled(self, enabled: bool) -> None:
        """Set the enabled state of the widget to `enabled`."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_parent(self) -> Widget:
        """Return the parent widget of this widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_parent(self, widget: Widget) -> None:
        """Set the parent widget of this widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_native_widget(self) -> Any:
        """Return the native backend widget instance.

        This is generally the widget that has the layout.
        """
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_root_native_widget(self) -> Any:
        """Return the root native backend widget.

        In most cases, this is the same as ``_mgui_get_native_widget``.  However, in
        cases where the native widget is in a scroll layout, this might be different.
        """
        raise NotImplementedError()

    @abstractmethod
    def _mgui_bind_parent_change_callback(
        self, callback: Callable[[Any], None]
    ) -> None:
        """Bind callback to parent change event."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_render(self) -> np.ndarray:
        """Return an RGBA (MxNx4) numpy array bitmap of the rendered widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_width(self) -> int:
        """Get the width of the widget.

        The intention is to get the width of the widget after it is shown, for the
        purpose of unifying widget width in a layout. Backends may do what they need to
        accomplish this. For example, Qt can use ``sizeHint().width()``, since
        ``width()`` may return something large if the widget has not yet been painted
        on screen.
        """
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_width(self, value: int) -> None:
        """Set the width of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_min_width(self) -> int:
        """Get the minimum width of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_min_width(self, value: int) -> None:
        """Set the minimum width of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_max_width(self) -> int:
        """Get the maximum width of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_max_width(self, value: int) -> None:
        """Set the maximum width of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_height(self) -> int:
        """Get the height of the widget.

        The intention is to get the height of the widget after it is shown, for the
        purpose of unifying widget height in a layout. Backends may do what they need to
        accomplish this. For example, Qt can use ``sizeHint().height()``, since
        ``height()`` may return something large if the widget has not yet been painted
        on screen.
        """
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_height(self, value: int) -> None:
        """Set the height of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_min_height(self) -> int:
        """Get the minimum height of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_min_height(self, value: int) -> None:
        """Set the minimum height of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_max_height(self) -> int:
        """Get the maximum height of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_max_height(self, value: int) -> None:
        """Set the maximum height of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_tooltip(self) -> str:
        """Get the tooltip for this widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_tooltip(self, value: str | None) -> None:
        """Set a tooltip for this widget."""
        raise NotImplementedError()


@runtime_checkable
class ValueWidgetProtocol(WidgetProtocol, Protocol):
    """Widget that has a current value, with getter/setter and on_change callback.

    It is worth noting that the **widget** is the thing that has a value.  Magicgui
    does not maintain & synchronize an independent model.
    """

    @abstractmethod
    def _mgui_get_value(self) -> Any:
        """Get current value of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_value(self, value: Any) -> None:
        """Set current value of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_bind_change_callback(self, callback: Callable[[Any], Any]) -> None:
        """Bind callback to value change event."""
        raise NotImplementedError()


@runtime_checkable
class SupportsReadOnly(Protocol):
    """Widget that can be read_only."""

    @abstractmethod
    def _mgui_set_read_only(self, value: bool) -> None:
        """Set read_only."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_read_only(self) -> bool:
        """Get read_only status."""
        raise NotImplementedError()


@runtime_checkable
class TableWidgetProtocol(ValueWidgetProtocol, SupportsReadOnly, Protocol):
    """ValueWidget subclass intended for 2D tabular data, with row & column headers."""

    @abstractmethod
    def _mgui_get_row_count(self) -> int:
        """Get the number of rows in the table."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_row_count(self, nrows: int) -> None:
        """Set the number of rows in the table. (Create/delete as needed)."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_remove_row(self, row: int) -> None:
        """Remove row at index `row`."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_column_count(self) -> int:
        """Get the number of columns in the table."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_column_count(self, ncols: int) -> None:
        """Set the number of columns in the table. (Create/delete as needed)."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_remove_column(self, column: int) -> None:
        """Remove column at index `column`."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_cell(self, row: int, col: int) -> Any:
        """Get current value of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_cell(self, row: int, col: int, value: Any) -> None:
        """Set current value of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_row_headers(self) -> tuple:
        """Get current row headers of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_row_headers(self, headers: Sequence) -> None:
        """Set current row headers of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_column_headers(self) -> tuple:
        """Get current column headers of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_column_headers(self, headers: Sequence) -> None:
        """Set current column headers of the widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_bind_row_headers_change_callback(
        self, callback: Callable[[Any], None]
    ) -> None:
        """Bind callback to row headers change event."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_bind_column_headers_change_callback(
        self, callback: Callable[[Any], None]
    ) -> None:
        """Bind callback to column headers change event."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_bind_change_callback(self, callback: Callable[[Any], Any]) -> None:
        """Bind callback to value change event."""
        raise NotImplementedError()


# note that "float" type hints also accept ints
# https://www.python.org/dev/peps/pep-0484/#the-numeric-tower
@runtime_checkable
class RangedWidgetProtocol(ValueWidgetProtocol, Protocol):
    """Value widget that supports numbers within a provided min/max range."""

    @abstractmethod
    def _mgui_get_min(self) -> float:
        """Get the minimum possible value."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_min(self, value: float) -> None:
        """Set the minimum possible value."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_max(self) -> float:
        """Get the maximum possible value."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_max(self, value: float) -> None:
        """Set the maximum possible value."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_step(self) -> float:
        """Get the step size."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_step(self, value: float) -> None:
        """Set the step size."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_adaptive_step(self) -> bool:
        """Get adaptive step status."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_adaptive_step(self, value: bool) -> None:
        """Set adaptive step status."""
        raise NotImplementedError()


@runtime_checkable
class SupportsChoices(Protocol):
    """Widget that has a set of valid choices."""

    @abstractmethod
    def _mgui_get_choices(self) -> tuple[tuple[str, Any], ...]:
        """Get available choices."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_choices(self, choices: Iterable[tuple[str, Any]]) -> None:
        """Set available choices."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_current_choice(self) -> str:
        """Return the text of the currently selected choice."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_choice(self, choice_name: str) -> Any:
        """Get data for a single choice."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_choice(self, choice_name: str, data: Any) -> None:
        """Set data for choice_name, or add a new item if choice_name doesn't exist."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_del_choice(self, choice_name: str) -> None:
        """Delete the provided ``choice_name`` and associated data."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_count(self) -> int:
        """Return number of choices."""
        raise NotImplementedError


@runtime_checkable
class CategoricalWidgetProtocol(ValueWidgetProtocol, SupportsChoices, Protocol):
    """Categorical widget, that has a set of valid choices, and a current value.

    It adds no additional methods.
    """


@runtime_checkable
class SupportsText(Protocol):
    """Widget that have text (in addition to value)... like buttons."""

    @abstractmethod
    def _mgui_set_text(self, value: str) -> None:
        """Set text."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_text(self) -> str:
        """Get text."""
        raise NotImplementedError()


@runtime_checkable
class SupportsIcon(Protocol):
    """Widget that can be reoriented."""

    @abstractmethod
    def _mgui_set_icon(self, value: str | None, color: str | None) -> None:
        """Set icon.

        Value is an "prefix:name" from iconify: https://icon-sets.iconify.design
        Color is any valid CSS color string.
        Set value to `None` or an empty string to remove icon.
        """


@runtime_checkable
class ButtonWidgetProtocol(ValueWidgetProtocol, SupportsText, SupportsIcon, Protocol):
    """The "value" in a ButtonWidget is the current (checked) state."""


@runtime_checkable
class SupportsOrientation(Protocol):
    """Widget that can be reoriented."""

    @abstractmethod
    def _mgui_set_orientation(self, value: str) -> None:
        """Set orientation, value will be 'horizontal' or 'vertical'."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_orientation(self) -> str:
        """Get orientation, return either 'horizontal' or 'vertical'."""
        raise NotImplementedError()


@runtime_checkable
class SliderWidgetProtocol(RangedWidgetProtocol, SupportsOrientation, Protocol):
    """Protocol for implementing a slider widget."""

    def _mgui_set_readout_visibility(self, visible: bool) -> None:
        """Set visibility of readout widget."""

    def _mgui_get_tracking(self) -> bool:
        """If tracking is False, changed is only emitted when released."""

    def _mgui_set_tracking(self, tracking: bool) -> None:
        """If tracking is False, changed is only emitted when released."""


# CONTAINER ----------------------------------------------------------------------


class ContainerProtocol(WidgetProtocol, SupportsOrientation, Protocol):
    """Widget that can contain other widgets.

    This generally manages a backend Layout.
    """

    @abstractmethod
    def _mgui_insert_widget(self, position: int, widget: Widget) -> None:
        """Insert `widget` at the given `position` in the layout."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_remove_widget(self, widget: Widget) -> None:
        """Remove the specified widget."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_get_margins(self) -> tuple[int, int, int, int]:
        """Get the margins of the container."""
        raise NotImplementedError()

    @abstractmethod
    def _mgui_set_margins(self, margins: tuple[int, int, int, int]) -> None:
        """Set the margins of the container."""
        raise NotImplementedError()


@runtime_checkable
class ToolBarProtocol(WidgetProtocol, Protocol):
    """Toolbar that contains a set of controls."""

    @abstractmethod
    def _mgui_add_button(
        self, text: str, icon: str, callback: Callable | None = None
    ) -> None:
        """Add a button to the toolbar."""

    @abstractmethod
    def _mgui_add_separator(self) -> None:
        """Add a separator line to the toolbar."""

    @abstractmethod
    def _mgui_add_spacer(self) -> None:
        """Add a spacer to the toolbar."""

    @abstractmethod
    def _mgui_add_widget(self, widget: Widget) -> None:
        """Add a widget to the toolbar."""

    @abstractmethod
    def _mgui_get_icon_size(self) -> tuple[int, int] | None:
        """Return the icon size of the toolbar."""

    @abstractmethod
    def _mgui_set_icon_size(self, size: int | tuple[int, int] | None) -> None:
        """Set the icon size of the toolbar."""

    @abstractmethod
    def _mgui_clear(self) -> None:
        """Clear the toolbar."""


class DialogProtocol(ContainerProtocol, Protocol):
    """Protocol for modal (blocking) containers."""

    @abstractmethod
    def _mgui_exec(self) -> None:
        """Show the dialog and block."""
        raise NotImplementedError()


class MainWindowProtocol(ContainerProtocol, Protocol):
    """Application main widget."""

    @abstractmethod
    def _mgui_create_menu_item(
        self,
        menu_name: str,
        action_name: str,
        callback: Callable | None = None,
        shortcut: str | None = None,
    ) -> None:
        """Create a new menu item.

        Parameters
        ----------
        menu_name : str
            The name of the menu to add the item to.
        action_name : str
            The name of the action to add.
        callback : Callable | None, optional
            A callback to be called when the action is triggered, by default None.
        shortcut : str | None, optional
            A keyboard shortcut for the action, by default None.
        """
        raise NotImplementedError()


# APPLICATION --------------------------------------------------------------------


class BaseApplicationBackend(ABC):
    """Backend Application object.

    Abstract class that provides an interface between backends and `Application`.
    Each backend must implement a subclass of `BaseApplicationBackend`, and
    implement all of its `_mgui_xxx` methods.
    """

    @abstractmethod
    def _mgui_get_backend_name(self) -> str:
        """Return the name of the backend."""

    @abstractmethod
    def _mgui_process_events(self) -> None:
        """Process all pending GUI events."""

    @abstractmethod
    def _mgui_run(self) -> None:
        """Start the application."""

    @abstractmethod
    def _mgui_quit(self) -> None:
        """Quit the native GUI event loop."""

    @abstractmethod
    def _mgui_get_native_app(self) -> Any:
        """Return the native GUI application instance."""

    @abstractmethod
    def _mgui_start_timer(
        self,
        interval: int = 0,
        on_timeout: Callable[[], None] | None = None,
        single: bool = False,
    ) -> None:
        """Create and start a timer.

        Parameters
        ----------
        interval : int, optional
            Interval between timeouts, by default 0
        on_timeout : Optional[Callable[[], None]], optional
            Function to call when timer finishes, by default None
        single : bool, optional
            Whether the timer should only fire once, by default False
        """

    @abstractmethod
    def _mgui_stop_timer(self) -> None:
        """Stop timer.  Should check for the existence of the timer."""



from __future__ import annotations

import inspect
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Literal,
    TypeVar,
    cast,
    overload,
)

from magicgui.widgets import FunctionGui, MainFunctionGui

if TYPE_CHECKING:
    from typing_extensions import ParamSpec

    from magicgui.application import AppRef

    _P = ParamSpec("_P")

__all__ = ["magicgui", "magic_factory", "MagicFactory"]


_R = TypeVar("_R")
_FGuiVar = TypeVar("_FGuiVar", bound=FunctionGui)


@overload
def magicgui(
    function: Callable[_P, _R],
    *,
    layout: str = "horizontal",
    scrollable: bool = False,
    labels: bool = True,
    tooltips: bool = True,
    call_button: bool | str | None = None,
    auto_call: bool = False,
    result_widget: bool = False,
    main_window: Literal[False] = False,
    app: AppRef | None = None,
    persist: bool = False,
    raise_on_unknown: bool = False,
    **param_options: dict,
) -> FunctionGui[_P, _R]: ...


@overload
def magicgui(
    function: Literal[None] | None = None,
    *,
    layout: str = "horizontal",
    scrollable: bool = False,
    labels: bool = True,
    tooltips: bool = True,
    call_button: bool | str | None = None,
    auto_call: bool = False,
    result_widget: bool = False,
    main_window: Literal[False] = False,
    app: AppRef | None = None,
    persist: bool = False,
    raise_on_unknown: bool = False,
    **param_options: dict,
) -> Callable[[Callable[_P, _R]], FunctionGui[_P, _R]]: ...


@overload
def magicgui(
    function: Callable[_P, _R],
    *,
    layout: str = "horizontal",
    scrollable: bool = False,
    labels: bool = True,
    tooltips: bool = True,
    call_button: bool | str | None = None,
    auto_call: bool = False,
    result_widget: bool = False,
    main_window: Literal[True],
    app: AppRef | None = None,
    persist: bool = False,
    raise_on_unknown: bool = False,
    **param_options: dict,
) -> MainFunctionGui[_P, _R]: ...


@overload
def magicgui(
    function: Literal[None] | None = None,
    *,
    layout: str = "horizontal",
    scrollable: bool = False,
    labels: bool = True,
    tooltips: bool = True,
    call_button: bool | str | None = None,
    auto_call: bool = False,
    result_widget: bool = False,
    main_window: Literal[True],
    app: AppRef | None = None,
    persist: bool = False,
    raise_on_unknown: bool = False,
    **param_options: dict,
) -> Callable[[Callable[_P, _R]], MainFunctionGui[_P, _R]]: ...


def magicgui(
    function: Callable | None = None,
    *,
    layout: str = "vertical",
    scrollable: bool = False,
    labels: bool = True,
    tooltips: bool = True,
    call_button: bool | str | None = None,
    auto_call: bool = False,
    result_widget: bool = False,
    main_window: bool = False,
    app: AppRef | None = None,
    persist: bool = False,
    raise_on_unknown: bool = False,
    **param_options: dict,
) -> Callable | FunctionGui:
    """Return a [`FunctionGui`][magicgui.widgets.FunctionGui] for `function`.

    Parameters
    ----------
    function : Callable, optional
        The function to decorate.  Optional to allow bare decorator with optional
        arguments. by default ``None``
    layout : str, optional
        The type of layout to use. Must be `horizontal` or `vertical`
        by default "vertical".
    scrollable : bool, optional
        Whether to enable scroll bars or not. If enabled, scroll bars will
        only appear along the layout direction, not in both directions.
    labels : bool, optional
        Whether labels are shown in the widget. by default True
    tooltips : bool, optional
        Whether tooltips are shown when hovering over widgets. by default True
    call_button : bool or str, optional
        If ``True``, create an additional button that calls the original
        function when clicked.  If a ``str``, set the button text. If None (the
        default), it defaults to True when ``auto_call`` is False, and False
        otherwise.
    auto_call : bool, optional
        If ``True``, changing any parameter in either the GUI or the widget attributes
        will call the original function with the current settings. by default False
    result_widget : bool, optional
        Whether to display a LineEdit widget the output of the function when called,
        by default False
    main_window : bool
        Whether this widget should be treated as the main app window, with menu bar,
        by default False.
    app : magicgui.Application or str, optional
        A backend to use, by default ``None`` (use the default backend.)
    persist : bool, optional
        If `True`, when parameter values change in the widget, they will be stored to
        disk and restored when the widget is loaded again with ``persist = True``.
        Call ``magicgui._util.user_cache_dir()`` to get the default cache location.
        By default False.
    raise_on_unknown : bool, optional
        If ``True``, raise an error if magicgui cannot determine widget for function
        argument or return type. If ``False``, ignore unknown types. By default False.
    param_options : dict[str, dict]
        Any additional keyword arguments will be used as parameter-specific options.
        Keywords must match the name of one of the arguments in the function signature,
        and the value must be a dict of keyword arguments to pass to the widget
        constructor.

    Returns
    -------
    result : FunctionGui or Callable[[F], FunctionGui]
        If ``function`` is not ``None`` (such as when this is used as a bare decorator),
        returns a FunctionGui instance, which is a list-like container of autogenerated
        widgets corresponding to each parameter in the function.
        If ``function`` is ``None`` such as when arguments are provided like
        ``magicgui(auto_call=True)``, then returns a function that can be used as a
        decorator.

    Examples
    --------
    >>> @magicgui
    ... def my_function(a: int = 1, b: str = "hello"):
    ...     pass
    >>> my_function.show()
    >>> my_function.a.value == 1  # True
    >>> my_function.b.value = "world"
    """
    return _magicgui(
        function=function,
        layout=layout,
        scrollable=scrollable,
        labels=labels,
        tooltips=tooltips,
        call_button=call_button,
        auto_call=auto_call,
        result_widget=result_widget,
        main_window=main_window,
        app=app,
        persist=persist,
        raise_on_unknown=raise_on_unknown,
        param_options=param_options,
    )


@overload
def magic_factory(
    function: Callable[_P, _R],
    *,
    layout: str = "horizontal",
    scrollable: bool = False,
    labels: bool = True,
    tooltips: bool = True,
    call_button: bool | str | None = None,
    auto_call: bool = False,
    result_widget: bool = False,
    main_window: Literal[False] = False,
    app: AppRef | None = None,
    persist: bool = False,
    widget_init: Callable[[FunctionGui], None] | None = None,
    raise_on_unknown: bool = False,
    **param_options: dict,
) -> MagicFactory[FunctionGui[_P, _R]]: ...


@overload
def magic_factory(
    function: Literal[None] | None = None,
    *,
    layout: str = "horizontal",
    scrollable: bool = False,
    labels: bool = True,
    tooltips: bool = True,
    call_button: bool | str | None = None,
    auto_call: bool = False,
    result_widget: bool = False,
    main_window: Literal[False] = False,
    app: AppRef | None = None,
    persist: bool = False,
    widget_init: Callable[[FunctionGui], None] | None = None,
    raise_on_unknown: bool = False,
    **param_options: dict,
) -> Callable[[Callable[_P, _R]], MagicFactory[FunctionGui[_P, _R]]]: ...


@overload
def magic_factory(
    function: Callable[_P, _R],
    *,
    layout: str = "horizontal",
    scrollable: bool = False,
    labels: bool = True,
    tooltips: bool = True,
    call_button: bool | str | None = None,
    auto_call: bool = False,
    result_widget: bool = False,
    main_window: Literal[True],
    app: AppRef | None = None,
    persist: bool = False,
    widget_init: Callable[[FunctionGui], None] | None = None,
    raise_on_unknown: bool = False,
    **param_options: dict,
) -> MagicFactory[MainFunctionGui[_P, _R]]: ...


@overload
def magic_factory(
    function: Literal[None] | None = None,
    *,
    layout: str = "horizontal",
    scrollable: bool = False,
    labels: bool = True,
    tooltips: bool = True,
    call_button: bool | str | None = None,
    auto_call: bool = False,
    result_widget: bool = False,
    main_window: Literal[True],
    app: AppRef | None = None,
    persist: bool = False,
    widget_init: Callable[[FunctionGui], None] | None = None,
    raise_on_unknown: bool = False,
    **param_options: dict,
) -> Callable[[Callable[_P, _R]], MagicFactory[MainFunctionGui[_P, _R]]]: ...


def magic_factory(
    function: Callable | None = None,
    *,
    layout: str = "vertical",
    scrollable: bool = False,
    labels: bool = True,
    tooltips: bool = True,
    call_button: bool | str | None = None,
    auto_call: bool = False,
    result_widget: bool = False,
    main_window: bool = False,
    app: AppRef | None = None,
    persist: bool = False,
    widget_init: Callable[[FunctionGui], None] | None = None,
    raise_on_unknown: bool = False,
    **param_options: dict,
) -> Callable | MagicFactory:
    """Return a [`MagicFactory`][magicgui.type_map._magicgui.MagicFactory] for function.

    `magic_factory` is nearly identical to the [`magicgui`][magicgui.magicgui] decorator
    with the following differences:

    1. Whereas `magicgui` returns a `FunctionGui` instance, `magic_factory` returns a
       callable that returns a `FunctionGui` instance.  (Technically, it returns an
       instance of [`MagicFactory`][magicgui.type_map._magicgui.MagicFactory] which you
       behaves exactly like a [`functools.partial`][functools.partial]
       for a `FunctionGui` instance.)
    2. `magic_factory` adds a `widget_init` method: a callable that will be called
        immediately after the `FunctionGui` instance is created.  This can be used to
        add additional widgets to the layout, or to connect signals to the widgets.

    !!!important
        Whereas decorating a function with `magicgui` will **immediately** create a
        widget instance, `magic_factory` will **not** create a widget instance until the
        decorated object is called.  This is often what you want in a library, whereas
        `magicgui` is useful for rapid, interactive development.

    Parameters
    ----------
    function : Callable, optional
        The function to decorate.  Optional to allow bare decorator with optional
        arguments. by default ``None``
    layout : str, optional
        The type of layout to use. Must be `horizontal` or `vertical`
        by default "vertical".
    scrollable : bool, optional
        Whether to enable scroll bars or not. If enabled, scroll bars will
        only appear along the layout direction, not in both directions.
    labels : bool, optional
        Whether labels are shown in the widget. by default True
    tooltips : bool, optional
        Whether tooltips are shown when hovering over widgets. by default True
    call_button : bool or str, optional
        If ``True``, create an additional button that calls the original
        function when clicked.  If a ``str``, set the button text. If None (the
        default), it defaults to True when ``auto_call`` is False, and False
        otherwise.
    auto_call : bool, optional
        If ``True``, changing any parameter in either the GUI or the widget attributes
        will call the original function with the current settings. by default False
    result_widget : bool, optional
        Whether to display a LineEdit widget the output of the function when called,
        by default False
    main_window : bool
        Whether this widget should be treated as the main app window, with menu bar,
        by default False.
    app : magicgui.Application or str, optional
        A backend to use, by default ``None`` (use the default backend.)
    persist : bool, optional
        If `True`, when parameter values change in the widget, they will be stored to
        disk and restored when the widget is loaded again with ``persist = True``.
        Call ``magicgui._util.user_cache_dir()`` to get the default cache location.
        By default False.
    widget_init : callable, optional
        A function that will be called with the newly created widget instance as its
        only argument.  This can be used to customize the widget after it is created.
        by default ``None``.
    raise_on_unknown : bool, optional
        If ``True``, raise an error if magicgui cannot determine widget for function
        argument or return type. If ``False``, ignore unknown types. By default False.
    param_options : dict of dict
        Any additional keyword arguments will be used as parameter-specific widget
        options. Keywords must match the name of one of the arguments in the function
        signature, and the value must be a dict of keyword arguments to pass to the
        widget constructor.

    Returns
    -------
    result : MagicFactory or Callable[[F], MagicFactory]
        If ``function`` is not ``None`` (such as when this is used as a bare decorator),
        returns a MagicFactory instance.
        If ``function`` is ``None`` such as when arguments are provided like
        ``magic_factory(auto_call=True)``, then returns a function that can be used as a
        decorator.

    Examples
    --------
    >>> @magic_factory
    ... def my_function(a: int = 1, b: str = "hello"):
    ...     pass
    >>> my_widget = my_function()
    >>> my_widget.show()
    >>> my_widget.a.value == 1  # True
    >>> my_widget.b.value = "world"
    """
    return _magicgui(
        factory=True,
        function=function,
        layout=layout,
        scrollable=scrollable,
        labels=labels,
        tooltips=tooltips,
        call_button=call_button,
        auto_call=auto_call,
        result_widget=result_widget,
        main_window=main_window,
        app=app,
        persist=persist,
        widget_init=widget_init,
        raise_on_unknown=raise_on_unknown,
        param_options=param_options,
    )


MAGICGUI_PARAMS = inspect.signature(magicgui).parameters


# _R is the return type of the decorated function
# _T is the type of the FunctionGui instance (FunctionGui or MainFunctionGui)
class MagicFactory(partial, Generic[_FGuiVar]):
    """Factory function that returns a FunctionGui instance.

    While this can be used directly, (see example below) the preferred usage is
    via the [`magicgui.magic_factory`][magicgui.magic_factory] decorator.

    Examples
    --------
    >>> def func(x: int, y: str):
    ...     pass
    >>> factory = MagicFactory(function=func, labels=False)
    >>> # factory accepts all the same arguments as magicgui()
    >>> widget1 = factory(call_button=True)
    >>> # can also override magic_kwargs that were provided when creating the factory
    >>> widget2 = factory(auto_call=True, labels=True)
    """

    _widget_init: Callable[[_FGuiVar], None] | None = None
    # func here is the function that will be called to create the widget
    # i.e. it will be either the FunctionGui or MainFunctionGui class
    func: Callable[..., _FGuiVar]

    def __new__(
        cls,
        function: Callable,
        *args: Any,
        magic_class: type[_FGuiVar] = FunctionGui,  # type: ignore
        widget_init: Callable[[_FGuiVar], None] | None = None,
        **keywords: Any,
    ) -> MagicFactory:
        """Create new MagicFactory."""
        if function is None:
            raise TypeError(
                "MagicFactory missing required positional argument 'function'"
            )

        # we want function first for the repr
        keywords = {"function": function, **keywords}
        if widget_init is not None:
            if not callable(widget_init):
                raise TypeError(
                    f"'widget_init' must be a callable, not {type(widget_init)}"
                )
            if len(inspect.signature(widget_init).parameters) != 1:
                raise TypeError(
                    "'widget_init' must be a callable that accepts a single argument"
                )
        obj = super().__new__(cls, magic_class, *args, **keywords)
        obj._widget_init = widget_init
        return obj

    def __repr__(self) -> str:
        """Return string repr."""
        args = [
            f"{k}={v!r}"
            for (k, v) in self.keywords.items()
            if v not in (MAGICGUI_PARAMS[k].default, {})
        ]
        return f"MagicFactory({', '.join(args)})"

    # TODO: annotate args and kwargs here so that
    # calling a MagicFactory instance gives proper mypy hints
    def __call__(self, *args: Any, **kwargs: Any) -> _FGuiVar:
        """Call the wrapped _magicgui and return a FunctionGui."""
        if args:
            raise ValueError("MagicFactory instance only accept keyword arguments")

        factory_kwargs = self.keywords.copy()
        prm_options = factory_kwargs.pop("param_options", {})
        prm_options.update(
            {k: kwargs.pop(k) for k in list(kwargs) if k not in MAGICGUI_PARAMS}
        )
        widget = self.func(param_options=prm_options, **{**factory_kwargs, **kwargs})
        if self._widget_init is not None:
            self._widget_init(widget)
        return widget

    def __getattr__(self, name: str) -> Any:
        """Allow accessing FunctionGui attributes without mypy error."""

    @property
    def __name__(self) -> str:
        """Pass function name."""
        return getattr(self.keywords.get("function"), "__name__", "FunctionGui")


def _magicgui(
    function: Callable | None = None,
    factory: bool = False,
    widget_init: Callable | None = None,
    main_window: bool = False,
    **kwargs: Any,
) -> Callable:
    """Actual private magicui decorator.

    if factory is `True` will return a MagicFactory instance, that can be called
    to return a `FunctionGui` instance.  See docstring of ``magicgui`` for parameters.
    Otherwise, this will return a FunctionGui instance directly.
    """

    def inner_func(func: Callable) -> FunctionGui | MagicFactory:
        if not callable(func):
            raise TypeError("the first argument must be callable")

        magic_class = MainFunctionGui if main_window else FunctionGui

        if factory:
            return MagicFactory(
                func, magic_class=magic_class, widget_init=widget_init, **kwargs
            )
        # MagicFactory is unnecessary if we are immediately instantiating the widget,
        # so we shortcut that and just return the FunctionGui here.
        return cast(FunctionGui, magic_class(func, **kwargs))

    return inner_func if function is None else inner_func(function)



"""# Manual progress bar

Example of a progress bar being updated manually.

"""

from magicgui import magicgui
from magicgui.widgets import ProgressBar


@magicgui(call_button="tick", pbar={"min": 0, "step": 2, "max": 20, "value": 0})
def manual(pbar: ProgressBar, increment: bool = True):
    """Example of manual progress bar control."""
    if increment:
        pbar.increment()
    else:
        pbar.decrement()


manual.show(run=True)



"""These are the main Widgets provided by magicgui.

While the primary API is to use the :func:`@magicgui <magicgui.magicgui>` decorator,
one can also instantiate widgets directly using any of these objects.

The :func:`create_widget` function is a helper function that will auto-pick and
instantiate the appropriate widget subclass given the arguments (type, annotation)
to the function.

"""

from functools import partial

from ._concrete import (
    CheckBox,
    ComboBox,
    Container,
    DateEdit,
    DateTimeEdit,
    Dialog,
    EmptyWidget,
    FileEdit,
    FloatRangeSlider,
    FloatSlider,
    FloatSpinBox,
    Label,
    LineEdit,
    ListEdit,
    LiteralEvalLineEdit,
    LogSlider,
    MainWindow,
    Password,
    ProgressBar,
    PushButton,
    QuantityEdit,
    RadioButton,
    RadioButtons,
    RangeEdit,
    RangeSlider,
    Select,
    SliceEdit,
    Slider,
    SpinBox,
    TextEdit,
    TimeEdit,
    ToolBar,
    TupleEdit,
)
from ._dialogs import request_values, show_file_dialog
from ._function_gui import FunctionGui, MainFunctionGui
from ._image import Image
from ._table import Table
from .bases import Widget, create_widget

#: Aliases for compatibility with ipywidgets.  (WIP)
IntSlider = Slider
FloatLogSlider = LogSlider
IntText = SpinBox
BoundedIntText = SpinBox
FloatText = FloatSpinBox
BoundedFloatText = FloatSpinBox
ToggleButton = RadioButton
Checkbox = CheckBox
Dropdown = ComboBox
Text = LineEdit
Textarea = TextEdit
Combobox = ComboBox
DatePicker = DateTimeEdit
Box = Container
HBox = partial(Container, layout="horizontal")
VBox = partial(Container, layout="vertical")
Button = PushButton

__all__ = [
    "CheckBox",
    "ComboBox",
    "Container",
    "create_widget",
    "DateEdit",
    "DateTimeEdit",
    "Dialog",
    "EmptyWidget",
    "FileEdit",
    "FloatSlider",
    "FloatSpinBox",
    "FloatRangeSlider",
    "FunctionGui",
    "Image",
    "Label",
    "LineEdit",
    "ListEdit",
    "LiteralEvalLineEdit",
    "LogSlider",
    "MainFunctionGui",
    "MainWindow",
    "Password",
    "PushButton",
    "ProgressBar",
    "QuantityEdit",
    "RadioButton",
    "RadioButtons",
    "RangeEdit",
    "RangeSlider",
    "Select",
    "SliceEdit",
    "Slider",
    "SpinBox",
    "Table",
    "TextEdit",
    "TimeEdit",
    "ToolBar",
    "TupleEdit",
    "Widget",
    "show_file_dialog",
    "request_values",
]

del partial



import contextlib
import sys
from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, ClassVar
from unittest.mock import Mock

import psygnal
import pytest

from magicgui.schema._guiclass import (
    GuiClass,
    button,
    guiclass,
    is_guiclass,
    unbind_gui_from_instance,
)
from magicgui.widgets import Container, PushButton


def test_guiclass():
    """Test that the guiclass decorator works as expected."""
    mock = Mock()

    @guiclass
    class Foo:
        a: int = 1
        b: str = "bar"

        @button
        def func(self):
            mock(asdict(self))

        # example recommended for type checking
        if TYPE_CHECKING:
            gui: ClassVar[Container]
            events: ClassVar[psygnal.SignalGroup]

    foo = Foo()

    assert foo.a == 1
    assert foo.b == "bar"

    assert isinstance(foo.gui, Container)
    assert isinstance(foo.gui.func, PushButton)
    assert foo.gui.a.value == 1
    assert foo.gui.b.value == "bar"

    foo.gui.a.value = 3
    assert foo.a == 3

    foo.b = "baz"
    assert foo.gui.b.value == "baz"

    foo.func()
    mock.assert_called_once_with({"a": 3, "b": "baz"})
    assert is_guiclass(Foo)
    assert is_guiclass(foo)


def test_frozen_guiclass():
    """Test that the guiclass decorator works as expected."""

    with pytest.raises(ValueError, match="not support dataclasses with `frozen=True`"):

        @guiclass(frozen=True)
        class Foo:
            a: int = 1
            b: str = "bar"


def test_on_existing_dataclass():
    """Test that the guiclass decorator works on pre-existing dataclasses."""

    @guiclass
    @dataclass
    class Foo:
        a: int = 1
        b: str = "bar"

    foo = Foo()
    assert foo.a == 1
    assert foo.b == "bar"
    assert isinstance(foo.gui, Container)


@pytest.mark.skipif(
    sys.version_info < (3, 11), reason="weakref_slot are python3.11 or higher"
)
def test_slots_guiclass():
    """Test that the guiclass decorator works as expected."""

    psyg_v = tuple(int(x.split("r")[0]) for x in psygnal.__version__.split(".")[:3])
    old_psygnal = psyg_v < (0, 6, 1)

    # if you don't use weakref_slot, it will still work, but you'll get a warning
    # during signal connection on gui-creation
    @guiclass(slots=True, weakref_slot=True)
    class Foo:
        a: int = 1
        b: str = "bar"

    foo = Foo()

    with (
        pytest.warns(UserWarning, match="Please update psygnal")
        if old_psygnal
        else contextlib.nullcontext()
    ):
        gui = foo.gui
        # note that with slots=True, the gui is recreated on every access
        assert foo.gui is not gui

    assert isinstance(gui, Container)
    assert gui.a.value == 1
    foo.b = "baz"
    assert gui.b.value == "baz"
    gui.a.value = 3

    if old_psygnal:
        # no change :(
        assert foo.a == 1
        assert len(gui.a.changed._slots) == 2
    else:
        assert foo.a == 3
        assert len(gui.a.changed._slots) == 3

    unbind_gui_from_instance(gui, foo)
    assert len(gui.a.changed._slots) == 2
    del foo


def test_guiclass_as_class():
    # variant on @guiclass, using class instead of decorator
    class T2(GuiClass):
        x: int
        y: str = "hi"

        @button
        def foo(self):
            return asdict(self)

    t2 = T2(1)
    assert t2.x == 1
    assert t2.y == "hi"
    assert t2.gui.x.value == 1
    assert t2.gui.y.value == "hi"
    t2.gui.x.value = 3
    assert t2.x == 3
    t2.y = "baz"
    assert t2.gui.y.value == "baz"
    assert isinstance(t2.gui.foo, PushButton)
    assert t2.foo() == {"x": 3, "y": "baz"}


def test_path_update():
    """One off test for FileEdits... which weren't updating.

    (The deeper issue is that things like FileEdit don't subclass ValueWidget...)
    """

    from pathlib import Path

    @guiclass
    class MyGuiClass:
        a: Path = Path("blabla")

    obj = MyGuiClass()
    assert obj.gui.a.value.stem == "blabla"
    assert obj.a.stem == "blabla"

    obj.gui.a.value = "foo"
    assert obj.gui.a.value.stem == "foo"
    assert obj.a.stem == "foo"



"""Experimental features.

Note: these are not guaranteed to be stable. Names and parameters may change or be
removed in future releases. Use at your own risk.
"""

from .schema._guiclass import button, guiclass, is_guiclass

__all__ = ["guiclass", "button", "is_guiclass"]



"""# Input values dialog

A basic example of a user input dialog.

This will pause code execution until the user responds.

# ![Values input dialog](../../images/values_input.png){ width=50% }

```python linenums="1"
from magicgui.widgets import request_values

vals = request_values(
    age=int,
    name={"annotation": str, "label": "Enter your name:"},
    title="Hi, who are you?",
)
print(repr(vals))
```
"""

# %%



"""# Jupyter notebooks and magicgui

This example shows magicgui widgets embedded in a jupyter notebook.

The key function here is `use_app("ipynb")`.

You can also [get this example at github](https://github.com/pyapp-kit/magicgui/blob/main/docs/examples/notebooks/magicgui_jupyter.ipynb).

```python hl_lines="4-5"
import math
from enum import Enum

from magicgui import magicgui, use_app
use_app("ipynb")

class Medium(Enum):
    # Various media and their refractive indices.
    Glass = 1.520
    Oil = 1.515
    Water = 1.333
    Air = 1.0003


@magicgui(
    call_button="calculate", result_widget=True, layout='vertical', auto_call=True
)
def snells_law(aoi=1.0, n1=Medium.Glass, n2=Medium.Water, degrees=True):
    # Calculate the angle of refraction given two media and an angle of incidence.
    if degrees:
        aoi = math.radians(aoi)
    try:
        n1 = n1.value
        n2 = n2.value
        result = math.asin(n1 * math.sin(aoi) / n2)
        return round(math.degrees(result) if degrees else result, 2)
    except ValueError:  # math domain error
        return "TIR!"


snells_law
```
"""

# %%
# ![magicgui widget embedded in the jupyter notebook](../../images/jupyter_magicgui_widget.png)



"""Most code in this module is vendored from matplotlib.

License agreement for matplotlib versions 1.3.0 and later
=========================================================

1. This LICENSE AGREEMENT is between the Matplotlib Development Team
("MDT"), and the Individual or Organization ("Licensee") accessing and
otherwise using matplotlib software in source or binary form and its
associated documentation.

2. Subject to the terms and conditions of this License Agreement, MDT
hereby grants Licensee a nonexclusive, royalty-free, world-wide license
to reproduce, analyze, test, perform and/or display publicly, prepare
derivative works, distribute, and otherwise use matplotlib
alone or in any derivative version, provided, however, that MDT's
License Agreement and MDT's notice of copyright, i.e., "Copyright (c)
2012- Matplotlib Development Team; All Rights Reserved" are retained in
matplotlib  alone or in any derivative version prepared by
Licensee.

3. In the event Licensee prepares a derivative work that is based on or
incorporates matplotlib or any part thereof, and wants to
make the derivative work available to others as provided herein, then
Licensee hereby agrees to include in any such work a brief summary of
the changes made to matplotlib .

4. MDT is making matplotlib available to Licensee on an "AS
IS" basis.  MDT MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, MDT MAKES NO AND
DISCLAIMS ANY REPRESENTATION OR WARRANTY OF MERCHANTABILITY OR FITNESS
FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF MATPLOTLIB
WILL NOT INFRINGE ANY THIRD PARTY RIGHTS.

5. MDT SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF MATPLOTLIB
 FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR
LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING
MATPLOTLIB , OR ANY DERIVATIVE THEREOF, EVEN IF ADVISED OF
THE POSSIBILITY THEREOF.

6. This License Agreement will automatically terminate upon a material
breach of its terms and conditions.

7. Nothing in this License Agreement shall be deemed to create any
relationship of agency, partnership, or joint venture between MDT and
Licensee.  This License Agreement does not grant permission to use MDT
trademarks or trade name in a trademark sense to endorse or promote
products or services of Licensee, or any third party.

8. By copying, installing or otherwise using matplotlib ,
Licensee agrees to be bound by the terms and conditions of this License
Agreement.
"""

import logging
from functools import lru_cache
from typing import TYPE_CHECKING, Collection, Optional, Union

try:
    import numpy as np
except ImportError as e:  # pragma: no cover
    msg = "Numpy required to use images in magicgui: use `pip install magicgui[image]`"
    raise type(e)(msg) from e

_log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from pathlib import Path

    import PIL.Image


def _is_mpl_norm(norm):
    try:
        from matplotlib import colors
    except ImportError:
        return False
    else:
        return isinstance(norm, colors.Normalize)


def _is_mpl_cmap(cmap):
    try:
        from matplotlib import colors
    except ImportError:
        return False
    else:
        return isinstance(cmap, colors.Colormap)


def _is_pil_image(image):
    try:
        from PIL.Image import Image
    except ImportError:
        return False
    else:
        return isinstance(image, Image)


# NOT from mpl
class Colormap:
    """Colormap that relates intensity values to colors.

    This colormap class is borrowed from napari, not mpl, but
    the ``__call__`` function has a similar API (without alpha).
    magicgui is also compatible with mpl colormap instances.

    Attributes
    ----------
    colors : array, shape (N, 4)
        Data used in the colormap.
    controls : array, shape (N,) or (N+1,)
        Control points of the colormap.
    interpolation : str
        Colormap interpolation mode, either 'linear' or
        'zero'. If 'linear', ncontrols = ncolors (one
        color per control point). If 'zero', ncontrols
        = ncolors+1 (one color per bin).
    """

    def __init__(
        self,
        colors: Collection = [[0.0, 0.0, 0.0, 1.0], [1.0, 1.0, 1.0, 1.0]],
        controls: Collection = np.zeros((0, 4)),
        interpolation: str = "linear",
    ) -> None:
        self.interpolation = interpolation
        self._colors = np.atleast_2d(colors)
        self._controls = controls

        if len(self.controls) == 0:
            n_controls = len(self._colors) + int(self.interpolation == "nearest")
            self._controls = np.linspace(0, 1, n_controls)

    @property
    def colors(self):
        return self._colors

    @property
    def controls(self):
        return self._controls

    def __call__(self, values, bytes=False):
        values = np.atleast_1d(values)

        lut = self.colors
        if self.interpolation == "linear":
            # One color per control point
            cols = [np.interp(values, self.controls, lut[:, i]) for i in range(4)]
            cols = np.stack(cols, axis=-1)
        elif self.interpolation == "nearest":
            # One color per bin
            indices = np.clip(np.searchsorted(self.controls, values) - 1, 0, len(lut))
            cols = lut[indices.astype(np.int32)]
        else:
            raise ValueError("Unrecognized Colormap Interpolation Mode")

        if bytes:
            cols = (cols * 255).astype(np.uint8)

        return cols


def get_cmap(name):
    if isinstance(name, Colormap) or _is_mpl_cmap(name):
        return name
    elif isinstance(name, str):
        try:
            import matplotlib
            import matplotlib.cm
        except ImportError as e:
            raise ImportError(
                "must install matplotlib to specify colormaps by name"
            ) from e
        else:
            try:
                return matplotlib.colormaps[name]
            except AttributeError:
                return matplotlib.cm.get_cmap(name)
    return Colormap()


class Normalize:
    """
    A class which, when called, linearly normalizes data into the
    ``[0.0, 1.0]`` interval.
    """

    def __init__(self, vmin=None, vmax=None, clip=False):
        """
        Parameters
        ----------
        vmin, vmax : float or None
            If *vmin* and/or *vmax* is not given, they are initialized from the
            minimum and maximum value, respectively, of the first input
            processed; i.e., ``__call__(A)`` calls ``autoscale_None(A)``.

        clip : bool, default: False
            If ``True`` values falling outside the range ``[vmin, vmax]``,
            are mapped to 0 or 1, whichever is closer, and masked values are
            set to 1.  If ``False`` masked values remain masked.

            Clipping silently defeats the purpose of setting the over, under,
            and masked colors in a colormap, so it is likely to lead to
            surprises; therefore the default is ``clip=False``.

        Notes
        -----
        Returns 0 if ``vmin == vmax``.
        """
        self.vmin = vmin
        self.vmax = vmax
        self.clip = clip

    @staticmethod
    def process_value(value):
        """
        Homogenize the input *value* for easy and efficient normalization.

        *value* can be a scalar or sequence.

        Returns
        -------
        result : masked array
            Masked array with the same shape as *value*.
        is_scalar : bool
            Whether *value* is a scalar.

        Notes
        -----
        Float dtypes are preserved; integer types with two bytes or smaller are
        converted to np.float32, and larger types are converted to np.float64.
        Preserving float32 when possible, and using in-place operations,
        greatly improves speed for large arrays.
        """
        is_scalar = not np.iterable(value)
        if is_scalar:
            value = [value]
        dtype = np.min_scalar_type(value)
        if np.issubdtype(dtype, np.integer) or dtype.type is np.bool_:
            # bool_/int8/int16 -> float32; int32/int64 -> float64
            dtype = np.promote_types(dtype, np.float32)
        # ensure data passed in as an ndarray subclass are interpreted as
        # an ndarray. See issue #6622.
        mask = np.ma.getmask(value)
        data = np.asarray(value)
        result = np.ma.array(data, mask=mask, dtype=dtype, copy=True)
        return result, is_scalar

    def __call__(self, value, clip=None):
        """
        Normalize *value* data in the ``[vmin, vmax]`` interval into the
        ``[0.0, 1.0]`` interval and return it.

        Parameters
        ----------
        value
            Data to normalize.
        clip : bool
            If ``None``, defaults to ``self.clip`` (which defaults to
            ``False``).

        Notes
        -----
        If not already initialized, ``self.vmin`` and ``self.vmax`` are
        initialized using ``self.autoscale_None(value)``.
        """
        if clip is None:
            clip = self.clip

        result, is_scalar = self.process_value(value)

        self.autoscale_None(result)
        # Convert at least to float, without losing precision.
        (vmin,), _ = self.process_value(self.vmin)
        (vmax,), _ = self.process_value(self.vmax)
        if vmin == vmax:
            result.fill(0)  # Or should it be all masked?  Or 0.5?
        elif vmin > vmax:
            raise ValueError("minvalue must be less than or equal to maxvalue")
        else:
            if clip:
                mask = np.ma.getmask(result)
                result = np.ma.array(
                    np.clip(result.filled(vmax), vmin, vmax), mask=mask
                )
            # ma division is very slow; we can take a shortcut
            resdat = result.data
            resdat -= vmin
            resdat /= vmax - vmin
            result = np.ma.array(resdat, mask=result.mask, copy=False)
        if is_scalar:
            result = result[0]
        return result

    def autoscale(self, A):
        """Set *vmin*, *vmax* to min, max of *A*."""
        A = np.asanyarray(A)
        self.vmin = A.min()
        self.vmax = A.max()

    def autoscale_None(self, A):
        """If vmin or vmax are not set, use the min/max of *A* to set them."""
        A = np.asanyarray(A)
        if self.vmin is None and A.size:
            self.vmin = A.min()
        if self.vmax is None and A.size:
            self.vmax = A.max()

    def scaled(self):
        """Return whether vmin and vmax are set."""
        return self.vmin is not None and self.vmax is not None


class ScalarMappable:
    """
    A mixin class to map scalar data to RGBA.

    The ScalarMappable applies data normalization before returning RGBA colors
    from the given colormap.
    """

    def __init__(self, norm=None, cmap=None):
        """

        Parameters
        ----------
        norm : `matplotlib.colors.Normalize` (or subclass thereof)
            The normalizing object which scales data, typically into the
            interval ``[0, 1]``.
            If *None*, *norm* defaults to a *colors.Normalize* object which
            initializes its scaling based on the first data processed.
        cmap : str or `~matplotlib.colors.Colormap`
            The colormap used to map normalized data values to RGBA colors.
        """
        self._A = None
        self.norm = None  # So that the setter knows we're initializing.
        self.set_norm(norm)  # The Normalize instance of this ScalarMappable.
        self.cmap = None  # So that the setter knows we're initializing.
        self.set_cmap(cmap)  # The Colormap instance of this ScalarMappable.
        #: The last colorbar associated with this ScalarMappable. May be None.
        self.colorbar = None

    def to_rgba(self, x, bytes=False, norm=True):
        """Return a normalized rgba array corresponding to *x*.

        In the normal case, *x* is a 1-D or 2-D sequence of scalars, and
        the corresponding ndarray of rgba values will be returned,
        based on the norm and colormap set for this ScalarMappable.

        There is one special case, for handling images that are already
        rgb or rgba, such as might have been read from an image file.
        If *x* is an ndarray with 3 dimensions,
        and the last dimension is either 3 or 4, then it will be
        treated as an rgb or rgba array, and no mapping will be done.
        The array can be uint8, or it can be floating point with
        values in the 0-1 range; otherwise a ValueError will be raised.
        If it is a masked array, the mask will be ignored.
        If the last dimension is 3, the *alpha* kwarg (defaulting to 1)
        will be used to fill in the transparency.  If the last dimension
        is 4, the *alpha* kwarg is ignored; it does not
        replace the pre-existing alpha.  A ValueError will be raised
        if the third dimension is other than 3 or 4.

        In either case, if *bytes* is *False* (default), the rgba
        array will be floats in the 0-1 range; if it is *True*,
        the returned rgba array will be uint8 in the 0 to 255 range.

        If norm is False, no normalization of the input data is
        performed, and it is assumed to be in the range (0-1).

        """
        # First check for special case, image input:
        try:
            if x.ndim == 3:
                if x.shape[2] == 3:
                    alpha = 1
                    if x.dtype == np.uint8:
                        alpha = np.uint8(alpha * 255)
                    m, n = x.shape[:2]
                    xx = np.empty(shape=(m, n, 4), dtype=x.dtype)
                    xx[:, :, :3] = x
                    xx[:, :, 3] = alpha
                elif x.shape[2] == 4:
                    xx = x
                else:
                    raise ValueError("Third dimension must be 3 or 4")
                if xx.dtype.kind == "f":
                    if norm and (xx.max() > 1 or xx.min() < 0):
                        raise ValueError(
                            "Floating point image RGB values "
                            "must be in the 0..1 range."
                        )
                    if bytes:
                        xx = (xx * 255).astype(np.uint8)
                elif xx.dtype == np.uint8:
                    if not bytes:
                        xx = xx.astype(np.float32) / 255
                else:
                    raise ValueError(
                        "Image RGB array must be uint8 or "
                        f"floating point; found {xx.dtype}"
                    )
                return xx
        except AttributeError:
            # e.g., x is not an ndarray; so try mapping it
            pass

        # This is the normal case, mapping a scalar array:
        x = np.ma.asarray(x)
        if norm:
            x = self.norm(x)
        rgba = self.cmap(x, bytes=bytes)
        return rgba

    def get_cmap(self):
        """Return the `.Colormap` instance."""
        return self.cmap

    def get_clim(self):
        """Return the values (min, max) that are mapped to the colormap limits."""
        return self.norm.vmin, self.norm.vmax

    def set_clim(self, vmin=None, vmax=None):
        """Set the norm limits for image scaling.

        Parameters
        ----------
        vmin, vmax : float
             The limits.

             The limits may also be passed as a tuple (*vmin*, *vmax*) as a
             single positional argument.

             .. ACCEPTS: (vmin: float, vmax: float)
        """
        if vmax is None:
            try:
                vmin, vmax = vmin
            except (TypeError, ValueError):
                pass
        if vmin is not None:
            self.norm.vmin = vmin
        if vmax is not None:
            self.norm.vmax = vmax
        self.changed()

    def set_cmap(self, cmap):
        """Set the colormap for luminance data.

        Parameters
        ----------
        cmap : `.Colormap` or str or None
        """
        in_init = self.cmap is None
        cmap = get_cmap(cmap)
        self.cmap = cmap
        if not in_init:
            self.changed()  # Things are not set up properly yet.

    def set_norm(self, norm):
        """Set the normalization instance.

        Parameters
        ----------
        norm : `.Normalize` or None

        Notes
        -----
        If there are any colorbars using the mappable for this norm, setting
        the norm of the mappable will reset the norm, locator, and formatters
        on the colorbar to default.
        """
        if norm is not None and not (isinstance(norm, Normalize) or _is_mpl_norm(norm)):
            raise TypeError(
                "`norm` must be an instance of mpl.Normalize or magicgui.Normalize"
            )
        in_init = self.norm is None
        if norm is None:
            norm = Normalize()
        self.norm = norm
        if not in_init:
            self.changed()  # Things are not set up properly yet.

    def autoscale(self):
        """
        Autoscale the scalar limits on the norm instance using the
        current array
        """
        if self._A is None:
            raise TypeError("You must first set_array for mappable")
        self.norm.autoscale(self._A)
        self.changed()

    def autoscale_None(self):
        """
        Autoscale the scalar limits on the norm instance using the
        current array, changing only limits that are None
        """
        if self._A is None:
            raise TypeError("You must first set_array for mappable")
        self.norm.autoscale_None(self._A)
        self.changed()

    def changed(self):
        pass


class Image(ScalarMappable):
    def __init__(self, cmap=None, norm=None):
        super().__init__(norm, cmap)
        self._imcache = None

    def set_data(
        self,
        A: Union[str, "Path", "np.ndarray", "PIL.Image.Image"],
        format: Optional[str] = None,
    ):
        """Set the image array.

        Note that this function does *not* update the normalization used.

        Parameters
        ----------
        A : array-like or `PIL.Image.Image`
        """
        from pathlib import Path

        if isinstance(A, Path):
            A = str(A)
        if isinstance(A, str):
            A = imread(A, format=format)
        elif _is_pil_image(A):
            A = pil_to_array(A)  # Needed e.g. to apply png palette.

        if not isinstance(A, np.ndarray):
            raise TypeError(
                "Image data must be a string, Path, numpy array, or PIL.Image"
            )

        # self._A = cbook.safe_masked_invalid(A, copy=True)
        self._A = A.copy()

        if self._A.dtype != np.uint8 and not np.can_cast(
            self._A.dtype, float, "same_kind"
        ):
            raise TypeError(
                f"Image data of dtype {self._A.dtype} cannot be converted to " "float"
            )

        if self._A.ndim == 3 and self._A.shape[-1] == 1:
            # If just one dimension assume scalar and apply colormap
            self._A = self._A[:, :, 0]

        if not (self._A.ndim == 2 or self._A.ndim == 3 and self._A.shape[-1] in [3, 4]):
            raise TypeError(
                f"Invalid shape {self._A.shape} for image data. Data must be 2D "
                "(monochromatic), or 3D: MxNx3 (RGB) or MxNx4 (RGBA)"
            )

        if self._A.ndim == 3:
            # If the input data has values outside the valid range (after
            # normalisation), we issue a warning and then clip X to the bounds
            # - otherwise casting wraps extreme values, hiding outliers and
            # making reliable interpretation impossible.
            high = 255 if np.issubdtype(self._A.dtype, np.integer) else 1
            if self._A.min() < 0 or high < self._A.max():
                _log.warning(
                    "Clipping input data to the valid range for imshow with "
                    "RGB data ([0..1] for floats or [0..255] for integers)."
                )
                self._A = np.clip(self._A, 0, high)
            # Cast unsupported integer types to uint8
            if self._A.dtype != np.uint8 and np.issubdtype(self._A.dtype, np.integer):
                self._A = self._A.astype(np.uint8)

        self._imcache = None

    def changed(self):
        self._imcache = None

    def make_image(self):
        return self._make_image(self._A)

    def _make_image(self, A):
        """Normalize, rescale, and colormap the image *A*

        Returns
        -------
        image : (M, N, 4) uint8 array
            The RGBA image, resampled unless *unsampled* is True.
        x, y : float
            The upper left corner where the image should be drawn, in pixel
            space.
        trans : Affine2D
            The affine transformation from image to pixel space.
        """
        if A is None:
            raise RuntimeError(
                "You must first set the image " "array or the image attribute"
            )
        if A.size == 0:
            raise RuntimeError(
                "_make_image must get a non-empty image. "
                "Your Artist's draw method must filter before "
                "this method is called."
            )

        if self._imcache is None:
            self._imcache = self.to_rgba(A, bytes=True, norm=(A.ndim == 2))

        return self._imcache


def pil_to_array(pilImage):
    """Load a `PIL image`_ and return it as a numpy int array.

    .. _PIL image: https://pillow.readthedocs.io/en/latest/reference/Image.html

    Returns
    -------
    numpy.array

        The array shape depends on the image type:

        - (M, N) for grayscale images.
        - (M, N, 3) for RGB images.
        - (M, N, 4) for RGBA images.
    """
    if pilImage.mode in ["RGBA", "RGBX", "RGB", "L"]:
        # return MxNx4 RGBA, MxNx3 RBA, or MxN luminance array
        return np.asarray(pilImage)
    elif pilImage.mode.startswith("I;16"):
        # return MxN luminance array of uint16
        raw = pilImage.tobytes("raw", pilImage.mode)
        if pilImage.mode.endswith("B"):
            x = np.frombuffer(raw, ">u2")
        else:
            x = np.frombuffer(raw, "<u2")
        return x.reshape(pilImage.size[::-1]).astype("=u2")
    else:  # try to convert to an rgba image
        try:
            pilImage = pilImage.convert("RGBA")
        except ValueError as err:
            raise RuntimeError("Unknown image mode") from err
        return np.asarray(pilImage)  # return MxNx4 RGBA array


@lru_cache
def _get_ssl_context():
    try:
        import certifi
    except ImportError:
        return None
    import ssl

    return ssl.create_default_context(cafile=certifi.where())


def imread(fname, format=None):
    """Read an image from a file into an array.

    Parameters
    ----------
    fname : str or file-like
        The image file to read: a filename, a URL or a file-like object opened
        in read-binary mode.
    format : str, optional
        The image file format assumed for reading the data. If not
        given, the format is deduced from the filename.  If nothing can
        be deduced, PNG is tried.

    Returns
    -------
    `numpy.array`
        The image data. The returned array has shape

        - (M, N) for grayscale images.
        - (M, N, 4) for RGBA images. (RGB will be padded to RGBA)
    """
    # hide imports to speed initial import on systems with slow linkers
    from pathlib import Path
    from urllib import parse

    try:
        import PIL.Image
        import PIL.PngImagePlugin
    except ImportError as e:  # pragma: no cover
        msg = (
            f"{e}. To load images from files or urls `pip install pillow`, "
            "or use the image extra: `pip install magicgui[image]`"
        )
        raise type(e)(msg) from e

    if format is None:
        if isinstance(fname, str):
            parsed = parse.urlparse(fname)
            # If the string is a URL (Windows paths appear as if they have a
            # length-1 scheme), assume png.
            if len(parsed.scheme) > 1:
                ext = parsed.path.rsplit(".", maxsplit=1)[-1] or "png"
            else:
                ext = Path(fname).suffix.lower()[1:]
        elif hasattr(fname, "geturl"):  # Returned by urlopen().
            # We could try to parse the url's path and use the extension, but
            # returning png is consistent with the block above.  Note that this
            # if clause has to come before checking for fname.name as
            # urlopen("file:///...") also has a name attribute (with the fixed
            # value "<urllib response>").
            ext = "png"
        elif hasattr(fname, "name"):
            ext = Path(fname.name).suffix.lower()[1:]
        else:
            ext = "png"
    else:
        ext = format
    img_open = PIL.PngImagePlugin.PngImageFile if ext == "png" else PIL.Image.open
    if isinstance(fname, str):
        parsed = parse.urlparse(fname)
        if len(parsed.scheme) > 1:  # Pillow doesn't handle URLs directly.
            # hide imports to speed initial import on systems with slow linkers
            from urllib import request

            ssl_ctx = _get_ssl_context()
            if ssl_ctx is None:
                from warnings import warn

                warn(
                    "Could not get certifi ssl context, https may not work.",
                    stacklevel=2,
                )
            with request.urlopen(fname, context=ssl_ctx) as response:
                import io

                try:
                    response.seek(0)
                except (AttributeError, io.UnsupportedOperation):
                    response = io.BytesIO(response.read())
                return imread(response, format=ext)
    with img_open(fname) as image:
        return pil_to_array(image)



import datetime
import inspect
from enum import Enum
from pathlib import Path
from typing import List, Optional, Tuple
from unittest.mock import MagicMock, patch

import pytest
from typing_extensions import Annotated

from magicgui import magicgui, types, use_app, widgets
from magicgui.widgets import Container, request_values
from magicgui.widgets.bases import DialogWidget, ValueWidget
from tests import MyInt


# it's important that "qt" be last here, so that it's used for
# the rest of the tests
@pytest.fixture(scope="module", params=["ipynb", "qt"])
def backend(request):
    return request.param


# FIXME: this test needs to come before we start swapping backends between qt and ipynb
# in other tests, otherwise it causes a stack overflow in windows...
# I'm not sure why that is, but it likely means that switching apps mid-process is
# not a good idea.  This should be explored further and perhaps prevented... and
# testing might need to be reorganized to avoid this problem.
def test_bound_callable_catches_recursion():
    """Test that accessing widget.value raises an informative error message.

    (... rather than a recursion error)
    """

    # this should NOT raise here. the function should not be called greedily
    @magicgui(x={"bind": lambda x: x.value * 2})
    def f(x: int = 5):
        return x

    with pytest.raises(RuntimeError):
        assert f() == 10
    f.x.unbind()
    assert f() == 5

    # use `get_value` within the callback if you need to access widget.value
    f.x.bind(lambda x: x.get_value() * 4)
    assert f() == 20


@pytest.mark.parametrize(
    "WidgetClass",
    [
        getattr(widgets, n)
        for n in widgets.__all__
        if n
        not in (
            "Widget",
            "TupleEdit",
            "FunctionGui",
            "MainFunctionGui",
            "show_file_dialog",
            "request_values",
            "create_widget",
        )
    ],
)
def test_widgets(WidgetClass, backend):
    """Test that we can retrieve getters, setters, and signals for most Widgets."""
    app = use_app(backend)
    if not hasattr(app.backend_module, WidgetClass.__name__):
        pytest.skip(f"no {WidgetClass.__name__!r} in backend {backend!r}")
    wdg: widgets.Widget = WidgetClass()
    wdg.close()


expectations = (
    [{"value": 1}, widgets.SpinBox],
    [{"value": 1.0}, widgets.FloatSpinBox],
    [{"value": "hi"}, widgets.LineEdit],
    [{"value": "a", "options": {"choices": ["a", "b"]}}, widgets.Combobox],
    [{"value": 1, "widget_type": "Slider"}, widgets.Slider],
)


@pytest.mark.parametrize("kwargs, expect_type", expectations)
def test_create_widget(kwargs, expect_type):
    """Test that various values get turned into widgets."""
    wdg = widgets.create_widget(**kwargs)
    assert isinstance(wdg, expect_type)
    wdg.close()


expectations_annotation = (
    (int, widgets.SpinBox),
    (float, widgets.FloatSpinBox),
    (range, widgets.RangeEdit),
    (str, widgets.LineEdit),
    (bool, widgets.CheckBox),
    (slice, widgets.SliceEdit),
    (datetime.date, widgets.DateEdit),
    (datetime.time, widgets.TimeEdit),
    (datetime.datetime, widgets.DateTimeEdit),
)


@pytest.mark.parametrize("annotation, expected_type", expectations_annotation)
def test_create_widget_annotation(annotation, expected_type):
    wdg = widgets.create_widget(annotation=annotation)
    assert isinstance(wdg, expected_type)
    wdg.close()


def test_create_widget_annotation_overwrite_parrams():
    wdg1 = widgets.create_widget(annotation=widgets.ProgressBar)
    assert isinstance(wdg1, widgets.ProgressBar)
    assert wdg1.visible
    wdg2 = widgets.create_widget(
        annotation=Annotated[widgets.ProgressBar, {"visible": False}]
    )
    assert isinstance(wdg2, widgets.ProgressBar)
    assert not wdg2.visible


# fmt: off
class MyBadWidget:
    """INCOMPLETE widget implementation and will error."""
    def _mgui_close_widget(self): ...
    def _mgui_get_visible(self): ...
    def _mgui_set_visible(self): ...
    def _mgui_get_enabled(self): ...
    def _mgui_set_enabled(self, enabled): ...
    def _mgui_get_parent(self): ...
    def _mgui_set_parent(self, widget): ...
    def _mgui_get_native_widget(self): return MagicMock()
    def _mgui_get_root_native_widget(self): ...
    def _mgui_bind_parent_change_callback(self, callback): ...
    def _mgui_render(self): ...
    def _mgui_get_width(self): ...
    def _mgui_set_width(self, value: int): ...
    def _mgui_get_min_width(self): ...
    def _mgui_set_min_width(self, value: int): ...
    def _mgui_get_max_width(self): ...
    def _mgui_set_max_width(self, value: int): ...
    def _mgui_get_height(self): ...
    def _mgui_set_height(self, value: int): ...
    def _mgui_get_min_height(self): ...
    def _mgui_set_min_height(self, value: int): ...
    def _mgui_get_max_height(self): ...
    def _mgui_set_max_height(self, value: int): ...
    def _mgui_get_value(self): ...
    def _mgui_set_value(self, value): ...
    def _mgui_bind_change_callback(self, callback): ...
    def _mgui_get_tooltip(self, value): ...
    # def _mgui_set_tooltip(self, value): ...


class MyValueWidget(MyBadWidget):
    """Complete protocol implementation... should work."""
    def _mgui_set_tooltip(self, value): ...
# fmt: on


def test_custom_widget():
    """Test that create_widget works with arbitrary backend implementations."""
    # by implementing the ValueWidgetProtocol, magicgui will know to wrap the above
    # widget with a widgets._bases.ValueWidget
    with pytest.warns(UserWarning, match="must accept a `parent` Argument"):
        wdg = widgets.create_widget(1, widget_type=MyValueWidget)  # type:ignore
    assert isinstance(wdg, ValueWidget)
    wdg.close()


def test_custom_widget_fails():
    """Test that create_widget works with arbitrary backend implementations."""
    with pytest.raises(TypeError) as err:
        widgets.create_widget(1, widget_type=MyBadWidget)  # type: ignore
    assert "does not implement 'WidgetProtocol'" in str(err)
    assert "Missing methods: {'_mgui_set_tooltip'}" in str(err)


def test_extra_kwargs_error():
    """Test that unrecognized kwargs gives a FutureWarning."""
    with pytest.raises(TypeError) as wrn:
        widgets.Label(unknown_kwarg="hi")
    assert "unexpected keyword argument" in str(wrn)


def test_autocall_no_runtime_error():
    """Make sure changing a value doesn't cause an autocall infinite loop."""

    @magicgui(auto_call=True, result_widget=True)
    def func(input=1):
        return round(input, 4)

    func.input.value = 2


def test_basic_widget_attributes():
    """Basic test coverage for getting/setting attributes."""
    widget = widgets.create_widget(value=1, name="my_name")
    container = widgets.Container(labels=False)
    assert widget.enabled
    widget.enabled = False
    assert not widget.enabled

    assert not widget.visible
    widget.show()
    assert widget.visible

    assert widget.parent is None
    container.append(widget)
    assert widget.parent is container
    widget.parent = None
    assert widget.parent is None
    assert widget.label == "my name"
    widget.label = "A different label"
    assert widget.label == "A different label"

    assert widget.param_kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
    widget.param_kind = inspect.Parameter.KEYWORD_ONLY
    widget.param_kind = "positional_only"
    assert widget.param_kind == inspect.Parameter.POSITIONAL_ONLY
    with pytest.raises(KeyError):
        widget.param_kind = "not a proper param type"
    with pytest.raises(TypeError):
        widget.param_kind = 1

    assert repr(widget) == "SpinBox(value=1, annotation=None, name='my_name')"
    assert widget.options == {
        "max": 999,
        "min": 0,
        "step": None,
        "enabled": False,
        "visible": False,
    }
    widget.close()


def test_width_height():
    widget = widgets.create_widget(value=1, name="my_name")
    widget.show()
    assert widget.visible
    assert widget.width < 100

    widget.width = 150.01
    assert widget.width == 150
    widget.min_width = 100.01
    assert widget.min_width == 100
    widget.max_width = 200.01
    assert widget.max_width == 200

    widget.height = 150.01
    assert widget.height == 150
    widget.min_height = 100.01
    assert widget.min_height == 100
    widget.max_height = 200.01
    assert widget.max_height == 200


def test_tooltip():
    label = widgets.Label()
    assert not label.tooltip
    label.tooltip = "My Tooltip"
    assert label.tooltip == "My Tooltip"


def test_widget_resolves_forward_ref():
    """The annotation on a widget should always be a resolved type."""

    @magicgui
    def widget(x: "tests.MyInt"):  # type: ignore  # noqa
        pass

    assert widget.x.annotation is MyInt


def test_unhashable_choice_data():
    """Test that providing unhashable choice data is ok."""
    combo = widgets.ComboBox()
    assert not combo.choices
    combo.choices = ("a", "b", "c")
    assert combo.choices == ("a", "b", "c")
    combo.choices = (("a", [1, 2, 3]), ("b", [1, 2, 5]))
    assert combo.choices == ([1, 2, 3], [1, 2, 5])
    combo.choices = ("x", "y", "z")
    assert combo.choices == ("x", "y", "z")
    combo.close()


def test_ambiguous_eq_choice_data():
    """Test that providing choice data with an ambiguous equal operation is ok."""
    import numpy as np

    combo = widgets.ComboBox()
    assert not combo.choices
    combo.choices = (("a", np.array([0, 0, 0])), ("b", np.array([1, 2, 3])))
    assert len(combo.choices) == 2
    assert np.all(combo.choices[0] == [0, 0, 0])
    assert np.all(combo.choices[1] == [1, 2, 3])

    combo.close()


def test_bound_values():
    """Test that we can bind a "permanent" value override to a parameter."""

    @magicgui(x={"bind": 10})
    def f(x: int = 5):
        return x

    # bound values hide the widget by default
    assert not f.x.visible
    assert f() == 10
    f.x.unbind()
    assert f() == 5


def test_bound_unknown_type_annotation():
    """Test that we can bind a "permanent" value override to a parameter."""

    import numpy as np

    def _provide_value(_):
        return np.array(1)

    @magicgui(arr={"bind": _provide_value})
    def f(arr: np.ndarray) -> np.ndarray:
        return arr

    assert f() == np.array(1)


def test_binding_None():
    """Test that we can bind a "permanent" value override to a parameter."""

    @magicgui(x={"bind": None})
    def f(x: int = 5):
        return x

    assert f() is None
    f.x.unbind()
    assert f() == 5


def test_bound_values_visible():
    """Test that we force a bound widget to be visible."""

    @magicgui(x={"bind": 10, "visible": True})
    def f(x: int = 5):
        return x

    f.show()
    assert f.x.visible
    assert f() == 10
    f.x.unbind()
    assert f() == 5


def test_bound_callables():
    """Test that we can use a callable as a bound value."""

    @magicgui(x={"bind": lambda x: 10})
    def f(x: int = 5):
        return x

    assert f() == 10
    f.x.unbind()
    assert f() == 5


def test_bound_callable_without_calling():
    """Test that we can use a callable as a bound value, but return it directly."""

    def callback():
        return "hi"

    @magicgui
    def f(x: int = 5):
        return x

    assert f() == 5
    f.x.bind(callback, call=False)
    assert f() == callback
    assert f()() == "hi"


def test_bound_not_called():
    """Test that"""
    mock = MagicMock()
    f = magicgui(lambda a: None, a={"bind": mock})
    # the bind function should not be called when creating the widget
    mock.assert_not_called()
    # the bind function should be called when getting the value
    _ = f.a.value
    mock.assert_called_once_with(f.a)


def test_progressbar():
    """Test manually controlling a progressbar."""

    @magicgui(pbar={"min": 20, "max": 40, "step": 2, "value": 30})
    def t(pbar: widgets.ProgressBar):
        assert pbar.get_value() == 30
        pbar.decrement()
        assert pbar.get_value() == 28
        pbar.step = 5
        assert pbar.get_value() == 28
        pbar.increment()
        assert pbar.get_value() == 33
        pbar.decrement(10)
        assert pbar.get_value() == 23
        return pbar.get_value()

    assert t() == 23


def test_main_function_gui():
    """Test that main_window makes the widget a top level main window with menus."""

    @magicgui(main_window=True)
    def add(num1: int, num2: int) -> int:
        """Adds the given two numbers, returning the result.

        The function assumes that the two numbers can be added and does
        not perform any prior checks.

        Parameters
        ----------
        num1 , num2 : int
            Numbers to be added

        Returns
        -------
        int
            Resulting integer
        """

    assert not add.visible
    add.show()
    assert add.visible

    assert isinstance(add, widgets.MainFunctionGui)
    add._show_docs()
    assert isinstance(add._help_text_edit, widgets.TextEdit)
    assert add._help_text_edit.value.startswith("Adds the given two numbers")
    assert add._help_text_edit.read_only
    add.close()


def test_range_widget():
    args = (-100, 1000, 2)
    rw = widgets.RangeEdit(*args)
    v = rw.value
    assert isinstance(v, range)
    assert (v.start, v.stop, v.step) == args


def test_range_widget_max():
    # max will override and restrict the possible values
    rw = widgets.RangeEdit(-100, 250, 1, max=(0, 500, 1))
    v = rw.value
    assert isinstance(v, range)
    assert (rw.start.max, rw.stop.max, rw.step.max) == (0, 500, 1)

    with pytest.raises(ValueError):
        rw = widgets.RangeEdit(100, 300, 5, max=(0, 500, 5))


def test_range_widget_min():
    # max will override and restrict the possible values
    rw = widgets.RangeEdit(2, 1000, 5, min=(0, 500, 5))
    v = rw.value
    assert isinstance(v, range)
    assert (rw.start.min, rw.stop.min, rw.step.min) == (0, 500, 5)

    with pytest.raises(ValueError):
        rw = widgets.RangeEdit(-100, 1000, 5, min=(0, 500, 5))


def test_range_value_none():
    """Test that arg: int = None defaults to 0"""

    @magicgui
    def f(x: Optional[int] = None):  # type: ignore
        ...

    assert f.x.value == 0
    rw = widgets.SpinBox(value=None)
    assert rw.value == 0


@pytest.mark.parametrize(
    "value,maksimum", [(10, 999), (None, 999), (1000, 9999), (1500, 9999)]
)
def test_range_big_value(value, maksimum):
    rw = widgets.SpinBox(value=value)
    assert rw.value == (value if value is not None else 0)
    rw.max = maksimum


def test_range_negative_value():
    rw = widgets.SpinBox(value=-10)
    assert rw.value == -10
    assert rw.min == -10


def test_adaptive():
    """Turn on and off adaptive step."""

    rw = widgets.SpinBox()
    assert rw.adaptive_step
    assert rw.step is None
    rw.adaptive_step = False
    assert not rw.adaptive_step
    assert rw.step == 1
    rw.step = None
    assert rw.adaptive_step
    assert rw.step is None
    rw.step = 3
    assert not rw.adaptive_step
    assert rw.step == 3

    rw = widgets.SpinBox(step=2)
    assert not rw.adaptive_step
    assert rw.step == 2
    rw.adaptive_step = True
    assert rw.adaptive_step
    assert rw.step is None
    rw.adaptive_step = False
    assert not rw.adaptive_step
    assert rw.step == 2


def test_exception_range_out_of_range():
    with pytest.raises(ValueError):
        widgets.SpinBox(value=10000, max=1000)

    with pytest.raises(ValueError):
        widgets.SpinBox(value=-10, min=0)


def test_file_dialog_events():
    """Test that file dialog events emit the value of the line_edit."""
    fe = widgets.FileEdit(value="hi")
    mock = MagicMock()
    fe.changed.connect(mock)
    fe.line_edit.value = "world"
    mock.assert_called_once_with(Path("world"))


def test_file_dialog_button_events():
    """Test that clicking the file dialog button doesn't emit an event."""
    fe = widgets.FileEdit(value="hi")
    mock = MagicMock()
    fe.changed.connect(mock)
    with patch.object(fe, "_show_file_dialog", return_value=""):
        fe.choose_btn.changed.emit("value")
    mock.assert_not_called()
    assert fe.value == Path("hi")


def test_file_edit_values():
    cwd = Path(".").absolute()

    fe = widgets.FileEdit(mode=types.FileDialogMode.EXISTING_FILE)
    assert isinstance(fe.value, Path)

    fe.value = Path("hi")
    assert fe.value == cwd / "hi"

    fe = widgets.FileEdit(mode=types.FileDialogMode.EXISTING_FILE, nullable=True)
    assert fe.value is None

    fe.value = Path("hi")
    assert fe.value == cwd / "hi"

    fe.value = None
    assert fe.value is None

    fe = widgets.FileEdit(mode=types.FileDialogMode.EXISTING_FILES)
    assert fe.value == ()

    fe.value = Path("hi")
    assert fe.value == (cwd / "hi",)

    fe.value = (Path("hi"), Path("world"))
    assert fe.value == (cwd / "hi", cwd / "world")

    fe.value = ()
    assert fe.value == ()


def test_null_events():
    """Test that nullable widgets emit events when their null value is set"""
    wdg = widgets.ComboBox(choices=["a", "b"], nullable=True)
    mock = MagicMock()
    wdg.changed.connect(mock)
    wdg.value = "b"
    mock.assert_called_once()
    mock.reset_mock()
    wdg.value = None
    mock.assert_called_once()
    mock.reset_mock()

    wdg._nullable = False
    wdg.value = "a"
    mock.assert_called_once()
    mock.reset_mock()
    mock.assert_not_called()
    wdg.value = None
    mock.assert_not_called()


@pytest.mark.parametrize("WdgClass", [widgets.FloatSlider, widgets.FloatSpinBox])
@pytest.mark.parametrize("value", [1, 1e6, 1e12, 1e16, 1e22])
def test_extreme_floats(WdgClass, value):
    wdg = WdgClass(value=value, max=value * 10)
    assert round(wdg.value / value, 4) == 1
    assert round(wdg.max / value, 4) == 10

    mock = MagicMock()
    wdg.changed.connect(mock)
    wdg.value = value * 2
    mock.assert_called_once()
    assert round(mock.call_args[0][0] / value, 4) == 2

    _value = 1 / value
    wdg2 = WdgClass(value=_value, step=_value / 10, max=_value * 100)
    assert round(wdg2.value / _value, 4) == 1.0
    wdg.close()
    wdg2.close()


@pytest.mark.parametrize("Cls", [widgets.ComboBox, widgets.RadioButtons])
def test_categorical_widgets(Cls):
    wdg = Cls(
        value=1,
        choices=[("first option", 1), ("second option", 2), ("third option", 3)],
    )

    mock = MagicMock()
    wdg.changed.connect(mock)
    assert isinstance(wdg, widgets.bases.CategoricalWidget)
    assert wdg.value == 1
    assert wdg.current_choice == "first option"
    mock.assert_not_called()
    wdg.value = 2
    mock.assert_called_once_with(2)
    assert wdg.value == 2
    assert wdg.current_choice == "second option"
    assert wdg.choices == (1, 2, 3)

    wdg.del_choice("third option")
    assert wdg.choices == (1, 2)


@pytest.mark.parametrize(
    "Cls,value", [(widgets.ComboBox, "c3"), (widgets.Select, ["c2", "c3"])]
)
def test_reset_choices_emits_once(Cls, value):
    data = ["c1", "c2", "c3"]
    wdg = Cls(
        value=value,
        choices=lambda w: data,
    )

    mock = MagicMock()
    wdg.changed.connect(mock)
    mock.assert_not_called()
    data = ["d2", "d4"]
    wdg.reset_choices()
    mock.assert_called_once()
    data = ["d2", "d4", "d5"]
    wdg.reset_choices()
    mock.assert_called_once()


@pytest.mark.parametrize(
    "Cls,value1,value2",
    [(widgets.ComboBox, "c4", "c1"), (widgets.Select, ["c3", "c4"], ["c1", "c2"])],
)
def test_set_value_emits_once(Cls, value1, value2):
    wdg = Cls(
        value=value1,
        choices=["c1", "c2", "c3", "c4"],
    )

    mock = MagicMock()
    wdg.changed.connect(mock)
    mock.assert_not_called()
    wdg.value = value2
    mock.assert_called_once()
    wdg.value = value2
    mock.assert_called_once()


@pytest.mark.parametrize(
    "Cls,value", [(widgets.ComboBox, "c3"), (widgets.Select, ["c3", "c4"])]
)
def test_set_choices_emits_once(Cls, value):
    wdg = Cls(
        value=value,
        choices=["c1", "c2", "c3", "c4"],
    )

    mock = MagicMock()
    wdg.changed.connect(mock)
    mock.assert_not_called()
    wdg.choices = ["d2", "d4", "d5"]
    mock.assert_called_once()


class MyEnum(Enum):
    A = "a"
    B = "b"


@pytest.mark.parametrize("Cls", [widgets.ComboBox, widgets.RadioButtons])
def test_categorical_widgets_with_enums(Cls):
    wdg = Cls(value=MyEnum.A, choices=MyEnum)

    mock = MagicMock()
    wdg.changed.connect(mock)
    assert isinstance(wdg, widgets.bases.CategoricalWidget)
    assert wdg.value == MyEnum.A
    assert wdg.current_choice == "A"
    mock.assert_not_called()
    wdg.value = MyEnum.B
    mock.assert_called_once_with(MyEnum.B)
    assert wdg.value == MyEnum.B
    assert wdg.current_choice == "B"
    assert wdg.choices == tuple(MyEnum.__members__.values())
    wdg.close()


@pytest.mark.parametrize("Cls", [widgets.ComboBox, widgets.RadioButtons])
def test_categorical_change_choices(Cls):
    """Make sure we can change choices to more or fewer options."""
    a = tuple(range(10))
    wdg = Cls(choices=a)
    assert wdg.choices == a
    b = tuple(range(5))
    wdg.choices = b
    assert wdg.choices == b
    c = tuple(range(15))
    wdg.choices = c
    assert wdg.choices == c


@pytest.mark.parametrize("Cls", [widgets.ComboBox, widgets.RadioButtons])
def test_categorical_change_choices_callable(Cls):
    first_choices = ("a", "b")

    def get_choices(wdg):
        return ("c", "d")

    wdg = Cls(choices=first_choices)
    assert wdg.choices == first_choices
    assert wdg._default_choices == first_choices

    wdg.choices = get_choices

    assert wdg.choices == ("c", "d")
    assert wdg._default_choices == get_choices


@pytest.mark.skipif(use_app().backend_name != "qt", reason="only on qt")
def test_radiobutton_reset_choices():
    """Test that reset_choices doesn't change the number of buttons."""
    from qtpy.QtWidgets import QRadioButton

    wdg = widgets.RadioButtons(choices=["a", "b", "c"])
    assert len(wdg.native.findChildren(QRadioButton)) == 3
    wdg.reset_choices()
    assert len(wdg.native.findChildren(QRadioButton)) == 3


def test_tracking():
    slider = widgets.Slider(tracking=False)
    assert slider.tracking is False
    slider.tracking = True
    assert slider.tracking


def test_select_set_value():
    sel = widgets.Select(value=[1, 3, 4], choices=list(range(10)))
    assert sel.value == [1, 3, 4]
    sel.value = [1, 4, 8]
    assert sel.value == [1, 4, 8]


def test_slider_readeout():
    """Test that the slider readout spinbox visibility works."""
    # FIXME: ugly direct backend access.
    sld = widgets.Slider()
    sld.show()
    backend_slider = sld.native.children()[1]
    assert "SpinBox" in type(backend_slider).__name__
    assert backend_slider.isVisible()

    sld = widgets.Slider(readout=False)
    sld.show()
    backend_slider = sld.native.children()[1]
    assert not backend_slider.isVisible()

    sld = widgets.Slider(readout=True)
    sld.show()
    assert sld.native.children()[1].isVisible()
    sld.readout = False
    assert not sld.native.children()[1].isVisible()


def test_slice_edit_events():
    """Test that changed events of spin boxes inside a slice edit are
    observable from its parent."""
    start, stop, step = 0, 10, 1
    sl = widgets.SliceEdit(start, stop, step)
    container = widgets.Container(widgets=[sl])
    mock = MagicMock()
    container.changed.connect(mock)
    sl.start.changed.emit(sl.value)
    mock.assert_called()
    assert sl.value == slice(start, stop, step)


def test_pushbutton_click_signal():
    btn = widgets.PushButton(text="click me")
    mock = MagicMock()
    mock2 = MagicMock()
    btn.changed.connect(mock)
    btn.clicked.connect(mock2)
    btn.native.click()
    mock.assert_called_once()
    mock2.assert_called_once()


def test_pushbutton_icon(backend: str):
    use_app(backend)
    btn = widgets.PushButton(icon="mdi:folder")
    btn.set_icon("play", "red")
    btn.set_icon(None)

    if backend == "qt":
        with pytest.warns(UserWarning, match="Could not set iconify icon"):
            btn.set_icon("bad:key")


def test_list_edit():
    """Test ListEdit."""
    mock = MagicMock()

    list_edit = widgets.ListEdit(value=[1, 2, 3])
    list_edit.changed.connect(mock)
    assert list_edit.value == [1, 2, 3]
    assert list_edit.data == [1, 2, 3]
    assert mock.call_count == 0

    list_edit.btn_plus.changed()
    assert list_edit.value == [1, 2, 3, 3]
    assert list_edit.data == [1, 2, 3, 3]
    assert mock.call_count == 1
    mock.assert_called_with([1, 2, 3, 3])

    list_edit[1].btn_minus.changed()
    assert list_edit.value == [1, 3, 3]
    assert list_edit.data == [1, 3, 3]
    assert mock.call_count == 2
    mock.assert_called_with([1, 3, 3])

    list_edit.data[0] = 0
    assert list_edit.value == [0, 3, 3]
    assert list_edit.data == [0, 3, 3]
    assert mock.call_count == 3
    mock.assert_called_with([0, 3, 3])

    list_edit[0].value = 10
    assert list_edit.value == [10, 3, 3]
    assert list_edit.data == [10, 3, 3]
    assert mock.call_count == 4
    mock.assert_called_with([10, 3, 3])

    list_edit.data[:2] = [6, 5]  # type: ignore
    assert list_edit.value == [6, 5, 3]
    assert list_edit.data == [6, 5, 3]
    assert mock.call_count == 5
    mock.assert_called_with([6, 5, 3])

    del list_edit.data[0]
    assert list_edit.value == [5, 3]
    assert list_edit.data == [5, 3]
    assert mock.call_count == 6
    mock.assert_called_with([5, 3])

    list_edit.value = [2, 1]
    assert list_edit.value == [2, 1]
    assert list_edit.data == [2, 1]
    # NOTE: changed.blocked() does not restore
    assert mock.call_count == 7
    mock.assert_called_with([2, 1])


def test_list_edit_only_values():
    @magicgui
    def f1(x=[2, 4, 6]):  # noqa: B006
        pass

    assert type(f1.x) is widgets.ListEdit
    assert f1.x._args_type is int
    assert f1.x.value == [2, 4, 6]

def test_list_edit_annotations():
    @magicgui
    def f2(x: List[int]):
        pass

    assert type(f2.x) is widgets.ListEdit
    assert f2.x.annotation == List[int]
    assert f2.x._args_type is int
    assert f2.x.value == []
    f2.x.btn_plus.changed()
    assert f2.x.value == [0]

    @magicgui(
        x={"options": {"widget_type": "Slider", "min": -10, "max": 10, "step": 5}}
    )
    def f3(x: List[int] = [0]):  # noqa: B006
        pass

    assert type(f3.x) is widgets.ListEdit
    assert type(f3.x[0].value_widget) is widgets.Slider
    assert f3.x[0].value_widget.min == -10
    assert f3.x[0].value_widget.max == 10
    assert f3.x[0].value_widget.step == 5

    @magicgui
    def f4(x: List[int] = ()):  # type: ignore
        pass

    assert type(f4.x) is widgets.ListEdit
    assert f4.x.annotation == List[int]
    assert f4.x._args_type is int
    assert f4.x.value == []
    f4.x.btn_plus.changed()
    assert type(f4.x[0].value_widget) is widgets.SpinBox
    assert f4.x.value == [0]

    @magicgui
    def f5(x: List[Annotated[int, {"max": 3}]]):
        pass

    assert type(f5.x) is widgets.ListEdit
    assert f5.x.annotation == List[int]
    f5.x.btn_plus.changed()
    assert f5.x[0].value_widget.max == 3


def test_tuple_edit():
    """Test TupleEdit."""
    from typing import Tuple

    mock = MagicMock()

    tuple_edit = widgets.TupleEdit(value=(1, "a", 2.5))
    tuple_edit.changed.connect(mock)
    assert tuple_edit.value == (1, "a", 2.5)
    assert mock.call_count == 0

    tuple_edit[0].value = 2
    assert tuple_edit.value == (2, "a", 2.5)
    assert mock.call_count == 1
    mock.assert_called_with((2, "a", 2.5))

    tuple_edit.value = (2, "xyz", 1.0)
    assert tuple_edit.value == (2, "xyz", 1.0)
    assert mock.call_count == 2
    mock.assert_called_with((2, "xyz", 1.0))

    with pytest.raises(ValueError):
        tuple_edit.value = (2, "x")

    @magicgui
    def f1(x=(2, 4, 6)):
        pass

    assert type(f1.x) is widgets.TupleEdit
    assert f1.x.value == (2, 4, 6)

    @magicgui
    def f2(x: Tuple[int, str]):
        pass

    assert type(f2.x) is widgets.TupleEdit
    assert f2.x.annotation == Tuple[int, str]
    assert f2.x.value == (0, "")

    @magicgui
    def f3(x: Tuple[Annotated[int, {"max": 3}], str]):
        pass

    assert type(f3.x) is widgets.TupleEdit
    assert f2.x.annotation == Tuple[int, str]
    assert f3.x[0].max == 3


def test_request_values(monkeypatch):
    from unittest.mock import Mock

    container = Container()

    mock = Mock()

    def _exec(self, **k):
        mock()
        assert self.native.parent() is container.native
        return True

    monkeypatch.setattr(DialogWidget, "exec", _exec)
    vals = request_values(
        age={"value": 40},
        name={"annotation": str, "label": "Enter your name:"},
        title="Hi! Who are you?",
        parent=container,
    )
    assert vals == {"age": 40, "name": ""}
    mock.assert_called_once()

    mock.reset_mock()
    vals = request_values(
        values={"age": int, "name": str}, title="Hi! Who are you?", parent=container
    )
    assert vals == {"age": 0, "name": ""}
    mock.assert_called_once()


def test_range_slider():
    @magicgui(auto_call=True, range_value={"widget_type": "RangeSlider", "max": 500})
    def func(range_value: Tuple[int, int] = (20, 380)):
        print(range_value)

    assert isinstance(func.range_value, widgets.RangeSlider)
    assert func.range_value.max == 500
    assert func.range_value.value == (20, 380)


def test_float_range_slider():
    @magicgui(auto_call=True, range_value={"widget_type": "FloatRangeSlider", "max": 1})
    def func(range_value: Tuple[float, float] = (0.2, 0.8)):
        print(range_value)

    assert isinstance(func.range_value, widgets.FloatRangeSlider)
    assert func.range_value.max == 1
    assert func.range_value.value == (0.2, 0.8)


def test_literal():
    from typing import Literal, Set

    from typing_extensions import get_args

    Lit = Literal[None, "a", 1, True, b"bytes"]

    @magicgui
    def f(x: Lit): ...

    cbox = f.x
    assert type(cbox) is widgets.ComboBox
    assert cbox.choices == get_args(Lit)

    @magicgui
    def f(x: Set[Lit]): ...

    sel = f.x
    assert type(sel) is widgets.Select
    assert sel.choices == get_args(Lit)


def test_separator_singleton():
    from magicgui.types import Separator, _Separator

    sep1 = Separator
    sep2 = _Separator()
    assert sep1 is sep2


def test_separator(backend: str):
    from magicgui.types import Separator

    use_app(backend)

    sep = [[Separator] * (i + 1) for i in range(4)]
    a = [1, 2, *sep[0], 4, *sep[1], 6, 7, *sep[2], 9, *sep[3], 11, 12, *sep[0], 14, *sep[0]]
    b = [1, 2, *sep[0], 4, *sep[1], 6, 7, *sep[2], 9, *sep[3], 6, 7, *sep[0], 9, *sep[0]]
    b2 = [1, 2, *sep[0], 4, *sep[1], 6, 7, *sep[2], 9, *sep[3], *sep[0], *sep[0]]

    combo_a = widgets.ComboBox(choices=a, value=a[0])
    combo_b = widgets.ComboBox(choices=b, value=b[0])

    assert len(combo_a) == len(combo_a.choices)
    assert len(combo_b) == len(combo_b.choices)

    if backend == "qt":
        def get_all_itemdata(combo_box):
            return [combo_box.itemData(index) for index in range(combo_box.count())]

        # Count returns the number of all items including separator items
        assert combo_a.native.count() == 21
        assert combo_b.native.count() == 18

        # Separator singletons themselves are used as separator item data
        assert get_all_itemdata(combo_a.native) == a
        assert get_all_itemdata(combo_b.native) == b2

        # Choices only returns unique, non-separator items
        assert combo_a.choices == (1, 2, 4, 6, 7, 9, 11, 12, 14)
        assert combo_b.choices == (1, 2, 4, 6, 7, 9)

    if backend == "ipynb":
        # Separator singletons themselves are used as separator item data
        assert combo_a.options['choices'] == a
        assert combo_b.options['choices'] == b  # items are not unique

        # Choices only returns duplicated, non-separator items
        assert combo_a.choices == (1, 2, 4, 6, 7, 9, 11, 12, 14)
        assert combo_b.choices == (1, 2, 4, 6, 7, 9, 6, 7, 9)


def test_float_slider_readout():
    sld = widgets.FloatSlider(value=4, min=0.5, max=10.5)
    assert sld.value == 4
    assert sld._widget._readout_widget.value() == 4
    assert sld._widget._readout_widget.minimum() == 0.5
    assert sld._widget._readout_widget.maximum() == 10.5


def test_toolbar():
    tb = widgets.ToolBar()
    tb.add_button("test", callback=lambda: None)
    tb.add_separator()
    tb.add_spacer()
    tb.add_button("test2", callback=lambda: None)
    tb.icon_size = 26
    assert tb.icon_size == (26, 26)
    tb.clear()



"""# Hotdog or not app

Demo app to upload an image and classify if it's an hotdog or not.
"""

import pathlib
from enum import Enum

from magicgui import magicgui


class HotdogOptions(Enum):
    """All hotdog possibilities."""

    Hotdog = 1
    NotHotdog = 0


@magicgui(main_window=True, layout="form", call_button="Classify", result_widget=True)
def is_hotdog(img: pathlib.Path) -> HotdogOptions:
    """Classify possible hotdog images.

    Upload an image and check whether it's an hotdog. For example, this image
    will be classified as one: <br><br>

    <img src="resources/hotdog.jpg">

    Parameters
    ----------
    img : pathlib.Path
        Path to a possible hotdog image

    Returns
    -------
    HotdogOptions
        True if image contains an hotdog in it
    """
    if "hotdog" in img.stem:
        return HotdogOptions.Hotdog
    return HotdogOptions.NotHotdog


if __name__ == "__main__":
    is_hotdog.show(run=True)



import sys
import types
import typing
from copy import copy
from functools import lru_cache, partial
from importlib import import_module
from typing import Any, Callable, Dict, Optional, Tuple, Type, Union

# required for python 3.8 to strip annotations
from typing_extensions import get_type_hints

try:
    from toolz import curry

    PARTIAL_TYPES: Tuple[Type, ...] = (partial, curry)
except ImportError:  # pragma: no cover
    PARTIAL_TYPES = (partial,)


@lru_cache(maxsize=1)
def _typing_names() -> Dict[str, Any]:
    return {**typing.__dict__, **types.__dict__}


def _unwrap_partial(func: Any) -> Any:
    while isinstance(func, PARTIAL_TYPES):
        func = func.func
    return func


def resolve_types(
    obj: Union[Callable, types.ModuleType, types.MethodType, type],
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
    do_imports: bool = False,
) -> Dict[str, Any]:
    """Resolve type hints from an object.

    Parameters
    ----------
    obj : Union[Callable, types.ModuleType, types.MethodType, type]
        Object to resolve type hints from.
    globalns : Optional[Dict[str, Any]]
        Global namespace to resolve imports from.
    localns : Optional[Dict[str, Any]]
        Local namespace to resolve imports from.
    do_imports : bool
        If `True`, will attempt to import modules when a NameError is
        encountered while resolving hints. For example, resolving `numpy.ndarray`
        will `import numpy` if a NameError is encountered.  By default, `False`.
    """
    # inject typing names into localns for convenience
    _localns = dict(_typing_names())
    # explicitly provided locals take precedence
    localns = {**_localns, **localns} if localns else _localns
    obj = _unwrap_partial(obj)

    try:
        hints = get_type_hints(obj, globalns=globalns, localns=localns)  # type: ignore
    except NameError as e:
        if do_imports:
            # try to import the top level name and try again
            msg = str(e)
            if msg.startswith("name ") and msg.endswith(" is not defined"):
                name = msg.split()[1].strip("\"'")
                ns = dict(localns) if localns else {}
                if name not in ns:
                    ns[name] = import_module(name)
                    return resolve_types(obj, globalns, ns, do_imports=do_imports)
        raise e

    # before version 3.9, typing.get_type_hints did not resolve hint.__args__
    if sys.version_info < (3, 9):
        _iterdict(hints, _resolve_forwards)
    return hints


def _iterdict(d: Dict[str, Any], func: Callable) -> None:
    for k, v in d.items():
        if isinstance(v, dict):
            _iterdict(v, func)
        else:
            d[k] = func(v)


def _resolve_forwards(v: Any) -> Any:
    if isinstance(v, typing.ForwardRef):
        return _try_cached_resolve(v.__forward_arg__)
    if getattr(v, "__args__", ()):
        v = copy(v)
        v.__args__ = tuple(_resolve_forwards(a) for a in v.__args__)
    return v


def resolve_single_type(
    hint: Any,
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
    do_imports: bool = True,
) -> Any:
    """Resolve a single type hint.

    See `resolve_types` for details on parameters.
    """
    if hint is None:
        return None
    mock_obj = type("_T", (), {"__annotations__": {"obj": hint}})()
    hints = resolve_types(mock_obj, globalns, localns, do_imports=do_imports)
    return hints["obj"]


_cached_resolve = lru_cache(maxsize=None)(resolve_single_type)


def _try_cached_resolve(v: Any) -> Any:
    try:
        return _cached_resolve(v)
    except TypeError:
        return resolve_single_type(v)



"""Types used internally in magicgui."""

from __future__ import annotations

from enum import Enum, EnumMeta
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Literal, Tuple, Union

from typing_extensions import TypedDict

if TYPE_CHECKING:
    from magicgui.widgets import FunctionGui
    from magicgui.widgets.bases import CategoricalWidget, Widget
    from magicgui.widgets.protocols import WidgetProtocol


class ChoicesDict(TypedDict):
    """Dict Type for setting choices in a categorical widget."""

    choices: ChoicesIterable
    key: Callable[[Any], str]


#: A [`Widget`][magicgui.widgets.Widget] class or a
#: [`WidgetProtocol`][magicgui.widgets.protocols.WidgetProtocol]
WidgetClass = Union["type[Widget]", "type[WidgetProtocol]"]
#: A generic reference to a :attr:`WidgetClass` as a string, or the class itself.
WidgetRef = Union[str, WidgetClass]
#: A :attr:`WidgetClass` (or a string representation of one) and a dict of kwargs
WidgetTuple = Tuple[WidgetRef, Dict[str, Any]]
#: An iterable that can be used as a valid argument for widget ``choices``
ChoicesIterable = Union[Iterable[Tuple[str, Any]], Iterable[Any]]
#: An callback that can be used as a valid argument for widget ``choices``.  It takes
#: a categorical widget and returns a :attr:`ChoicesIterable`.
ChoicesCallback = Callable[["CategoricalWidget[Any]"], ChoicesIterable]
#: The set of all valid types for widget ``choices``.
ChoicesType = Union[EnumMeta, ChoicesIterable, ChoicesCallback, ChoicesDict]
#: A callback that may be registered for a given return annotation. When called, it will
#: be provided an instance of a
#: [~magicgui.widgets.FunctionGui][magicgui.widgets.FunctionGui],
#: the result of the function that was called, and the return annotation itself.
ReturnCallback = Callable[["FunctionGui", Any, type], None]
#: A valid file path type
PathLike = Union[Path, str, bytes]


class FileDialogMode(Enum):
    """FileDialog mode options.

    - ``EXISTING_FILE`` - returns one existing file.
    - ``EXISTING_FILES`` - return one or more existing files.
    - ``OPTIONAL_FILE`` - return one file name that does not have to exist.
    - ``EXISTING_DIRECTORY`` - returns one existing directory.
    """

    EXISTING_FILE = "r"
    EXISTING_FILES = "rm"
    OPTIONAL_FILE = "w"
    EXISTING_DIRECTORY = "d"


class _Undefined:
    """Sentinel class to indicate the lack of a value when ``None`` is ambiguous.

    ``_Undefined`` is a singleton.
    """

    _singleton: _Undefined | None = None

    def __new__(cls) -> _Undefined:
        if _Undefined._singleton is None:
            _Undefined._singleton = super().__new__(cls)
        return _Undefined._singleton

    def __repr__(self) -> str:
        return "..."

    def __bool__(self) -> bool:
        return False


Undefined = _Undefined()


class _Separator:
    """Sentinel class to separate groups of items in a ComboBox.

    Example: ``choices=[1, 2, 3, Separator, Separator, 4, Separator]``

    ``_Separator`` is a singleton.
    """

    _instance: _Separator | None = None

    def __new__(cls) -> _Separator:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __repr__(self) -> str:
        return "<Separator>"


Separator = _Separator()


JsonStringFormats = Literal[
    # ISO 8601 format.
    # https://www.iso.org/iso-8601-date-and-time-format.html
    "date-time",  # 2018-11-13T20:20:39+00:00
    "time",  # 20:20:39+00:00, Draft 7
    "date",  # 2018-11-13, Draft 7
    "duration",  # P1Y2M3DT4H5M6S, Draft 2019-09
    # email addresses
    "email",
    "idn-email",  # Draft 7
    # hostnames
    "hostname",  # internet host name, RFC 1123
    "idn-hostname",
    # ip addresses
    "ipv4",
    "ipv6",
    # resource identifiers
    "uuid",  # 3e4666bf-d5e5-4aa7-b8ce-cefe41c7568a, draft 2019-09
    "uri",  # RFC3986
    "uri-reference",  # RFC3986, section 4.1., draft 6
    "iri",  # draft 7
    "iri-reference",  # draft 7
    # uri template
    "uri-template",  # RFC6570, draft 6
    # json pointer
    "json-pointer",  # RFC6901, draft 6
    "relative-json-pointer",  # draft 7
    # regular expressions
    "regex",  # draft 7
]



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from psygnal import Signal, SignalInstance

from magicgui.types import Undefined, _Undefined

from ._value_widget import ValueWidget

if TYPE_CHECKING:
    from typing_extensions import Unpack

    from magicgui.widgets import protocols

    from ._widget import WidgetKwargs


class ButtonWidget(ValueWidget[bool]):
    """Widget with a value, Wraps a widget implementing the ButtonWidgetProtocol.

    >see [ButtonWidgetProtocol][magicgui.widgets.protocols.ButtonWidgetProtocol].

    Parameters
    ----------
    value : bool
        The starting state of the widget.
    text : str, optional
        The text to display on the button. If not provided, will use ``name``.
    bind : Callable[[ValueWidget], Any] | Any, optional
        A value or callback to bind this widget. If provided, whenever
        [`widget.value`][magicgui.widgets.bases.ValueWidget.value] is
        accessed, the value provided here will be returned instead. `bind` may be a
        callable, in which case `bind(self)` will be returned (i.e. your bound callback
        must accept a single parameter, which is this widget instance).
    nullable : bool, optional
        If `True`, the widget will accepts `None` as a valid value, by default `False`.
    **base_widget_kwargs : Any
        All additional keyword arguments are passed to the base
        [`magicgui.widgets.Widget`][magicgui.widgets.Widget] constructor.
    """

    _widget: protocols.ButtonWidgetProtocol
    changed = Signal(
        object,
        description="Emitted when the button is clicked (may also be "
        "connected at the alias `clicked`).",
    )

    def __init__(
        self,
        value: bool | _Undefined = Undefined,
        *,
        text: str | None = None,
        icon: str | None = None,
        icon_color: str | None = None,
        bind: bool | Callable[[ValueWidget], bool] | _Undefined = Undefined,
        nullable: bool = False,
        **base_widget_kwargs: Unpack[WidgetKwargs],
    ) -> None:
        if text and base_widget_kwargs.get("label"):
            from warnings import warn

            warn(
                "'text' and 'label' are synonymous for button widgets. To suppress this"
                " warning, only provide one of the two kwargs.",
                stacklevel=2,
            )
        text = text or base_widget_kwargs.get("label")
        # TODO: make a backend hook that lets backends inject their optional API
        # ipywidgets button texts are called descriptions
        text = text or base_widget_kwargs.pop("description", None)
        super().__init__(
            value=value, bind=bind, nullable=nullable, **base_widget_kwargs
        )
        self.text = (text or self.name).replace("_", " ")
        if icon:
            self.set_icon(icon, icon_color)

    @property
    def options(self) -> dict:
        """Return options currently being used in this widget."""
        d = super().options.copy()
        d.update({"text": self.text})
        return d

    @property
    def text(self) -> str:
        """Text of the widget."""
        return self._widget._mgui_get_text()

    @text.setter
    def text(self, value: str) -> None:
        self._widget._mgui_set_text(value)

    @property
    def clicked(self) -> SignalInstance:
        """Alias for changed event."""
        return self.changed

    def set_icon(self, value: str | None, color: str | None = None) -> None:
        self._widget._mgui_set_icon(value, color)



from pathlib import Path

import pytest

from magicgui.widgets import Image

_mpl_image = pytest.importorskip("magicgui._mpl_image")
np = pytest.importorskip("numpy")
pilImage = pytest.importorskip("PIL.Image")


def test_image_widget():
    # png works
    image = Image(value=Path(__file__).parent / "_test.png")
    assert isinstance(image._image, _mpl_image.Image)
    assert isinstance(image.image_data, np.ndarray)
    assert isinstance(image.image_rgba, np.ndarray)
    assert image.image_data.shape == (200, 232, 3)
    assert image.image_rgba.shape == (200, 232, 4)

    prev = image.image_data.copy()
    # jpg works works
    image.value = Path(__file__).parent / "_test.jpg"
    # png and jpg different enough
    assert not np.allclose(image.image_data, prev)
    assert image.image_data.shape == (200, 232, 3)
    assert image.image_rgba.shape == (200, 232, 4)

    # 2D float
    image.value = np.random.rand(64, 64)
    assert image.image_data.shape == (64, 64)
    assert image.image_rgba.shape == (64, 64, 4)

    # 2D uint8
    image.value = np.random.randint(0, 255, (60, 60)).astype("uint8")
    assert image.image_data.shape == (60, 60)
    assert image.image_rgba.shape == (60, 60, 4)

    # RGBA float
    image.value = np.random.rand(60, 60, 4).astype("float")
    assert image.image_data.shape == (60, 60, 4)
    assert image.image_rgba.shape == (60, 60, 4)


def test_clim():
    # 2D uint8
    image = Image()
    image.value = np.random.rand(60, 60)
    rendered = image.image_rgba
    assert isinstance(rendered, np.ndarray)
    assert rendered.shape == (60, 60, 4)
    image.set_clim(0.5, 0.7)
    rendered2 = image.image_rgba
    assert not np.allclose(rendered, rendered2)
    assert image.get_clim() == (0.5, 0.7)


def test_empty_image():
    image = Image()
    assert image.image_data is None
    assert image.image_rgba is None
    assert image.get_clim() == (None, None)

    with pytest.raises(RuntimeError):
        image.set_clim(0, 100)
    with pytest.raises(RuntimeError):
        image.set_clim(0, 100)
    with pytest.raises(RuntimeError):
        image.set_cmap("turbo")
    with pytest.raises(RuntimeError):
        image.set_norm(1)


def test_pilImage():
    with pilImage.open(Path(__file__).parent / "_test.jpg") as img:
        image = Image(value=img)
    assert isinstance(image._image, _mpl_image.Image)
    assert isinstance(image.image_data, np.ndarray)
    assert image.image_data.shape == (200, 232, 3)
    assert image._image.make_image().shape == (200, 232, 4)


def test_internal_cmap():
    cmap = _mpl_image.Colormap([[0, 0, 0, 1], [1, 0, 0, 1]])
    image = Image()
    data = np.random.randint(0, 255, (60, 60)).astype("uint8")
    image.set_data(data, cmap=cmap)
    rendered = image.image_rgba
    assert isinstance(rendered, np.ndarray)
    assert rendered.shape == (60, 60, 4)

    image.set_cmap(
        _mpl_image.Colormap([[0, 0, 0, 1], [1, 0, 0, 1]], interpolation="nearest")
    )
    rendered2 = image.image_rgba
    assert isinstance(rendered2, np.ndarray)
    assert not np.allclose(rendered, rendered2)


def test_mpl_cmap():
    # 2D uint8
    image = Image()
    data = np.random.randint(0, 255, (60, 60)).astype("uint8")
    image.set_data(data)
    rendered = image.image_rgba
    assert isinstance(rendered, np.ndarray)
    # without a colormap, the output data should be nearly identical
    assert data.size - np.count_nonzero(rendered[..., 0] == data) < data.size * 0.01
    assert isinstance(rendered, np.ndarray)
    assert rendered.shape == (60, 60, 4)

    image.set_cmap("magma")
    rendered2 = image.image_rgba
    assert isinstance(rendered2, np.ndarray)
    # the colormap has been applied
    assert data.size - np.count_nonzero(rendered2[..., 0] == data) > 3000

    # can also use an mpl colormap instance
    image.set_cmap(_mpl_image.get_cmap("viridis"))
    rendered3 = image.image_rgba
    assert isinstance(rendered3, np.ndarray)
    # the colormap has been applied
    assert not np.allclose(rendered2, rendered3)


def test_mpl_norm():
    mplcolor = pytest.importorskip("matplotlib.colors")

    # 2D uint8
    image = Image()
    data = np.random.randint(0, 255, (60, 60)).astype("uint8")
    image.set_data(data)
    rendered = image.image_rgba
    assert isinstance(rendered, np.ndarray)
    assert rendered.shape == (60, 60, 4)

    image.set_norm(mplcolor.PowerNorm(0.5))
    rendered2 = image.image_rgba
    assert isinstance(rendered2, np.ndarray)
    assert rendered2.shape == (60, 60, 4)
    assert not np.allclose(rendered, rendered2)



import pytest

from magicgui.application import use_app


@pytest.fixture(scope="session")
def qapp():
    yield use_app("qt").native


# for now, the only backend is qt, and pytest-qt's qapp provides some nice pre-post
# test cleanup that prevents some segfaults.  Once we start testing multiple backends
# this will need to change.
@pytest.fixture(autouse=True, scope="function")
def always_qapp(qapp):
    yield qapp
    for w in qapp.topLevelWidgets():
        w.close()
        w.deleteLater()


@pytest.fixture(autouse=True, scope="function")
def _clean_return_callbacks():
    from magicgui.type_map._type_map import _RETURN_CALLBACKS

    yield

    _RETURN_CALLBACKS.clear()



"""# napari image arithmetic widget

[napari](https://github.com/napari/napari) is a fast, interactive,
multi-dimensional image viewer for python.  It uses Qt for the GUI, so it's easy
to extend napari with small, composable widgets created with `magicgui`.  Here
we're going to build this simple image arithmetic widget with a few additional
lines of code.

For napari-specific magicgui documentation, see the [napari
docs](https://napari.org/guides/magicgui.html)

![napari image arithmetic widget](../../images/imagemath.gif){ width=80% }

## outline

**This example demonstrates how to:**

1. Create a `magicgui` widget that can be used in another program (napari)

2. Use an `Enum` to create a dropdown menu

3. Connect some event listeners to create interactivity.

## code

*Code follows, with explanation below... You can also [get this example at
github](https://github.com/pyapp-kit/magicgui/blob/main/docs/examples/napari/napari_image_arithmetic.py).*

```python linenums="1" hl_lines="25 38"
from enum import Enum

import napari
import numpy
from napari.types import ImageData

from magicgui import magicgui


class Operation(Enum):
    # A set of valid arithmetic operations for image_arithmetic.
    #
    # To create nice dropdown menus with magicgui, it's best
    # (but not required) to use Enums.  Here we make an Enum
    # class for all of the image math operations we want to
    # allow.
    add = numpy.add
    subtract = numpy.subtract
    multiply = numpy.multiply
    divide = numpy.divide


# here's the magicgui!  We also use the additional
# `call_button` option
@magicgui(call_button="execute")
def image_arithmetic(
    layerA: ImageData, operation: Operation, layerB: ImageData
) -> ImageData:
    # Add, subtracts, multiplies, or divides to image layers.
    return operation.value(layerA, layerB)

# create a viewer and add a couple image layers
viewer = napari.Viewer()
viewer.add_image(numpy.random.rand(20, 20), name="Layer 1")
viewer.add_image(numpy.random.rand(20, 20), name="Layer 2")

# add our new magicgui widget to the viewer
viewer.window.add_dock_widget(image_arithmetic)

# keep the dropdown menus in the gui in sync with the layer model
viewer.layers.events.inserted.connect(image_arithmetic.reset_choices)
viewer.layers.events.removed.connect(image_arithmetic.reset_choices)

napari.run()
```

## walkthrough

We're going to go a little out of order so that the other code makes more sense.
Let's start with the actual function we'd like to write to do some image arithmetic.

### the function

Our function takes two `numpy` arrays (in this case, from [Image layers](https://napari.org/howtos/layers/image.html)),
and some mathematical operation
(we'll restrict the options using an `enum.Enum`).
When called, ourfunction calls the selected operation on the data.

```python
def image_arithmetic(array1, operation, array2):
    return operation.value(array1, array2)
```

#### type annotations

`magicgui` works particularly well with [type annotations](https://docs.python.org/3/library/typing.html),
and allows third-party libraries to register widgets and behavior for handling
their custom types (using [`magicgui.type_map.register_type`][magicgui.type_map.register_type]).
`napari` [provides support for `magicgui`](https://github.com/napari/napari/blob/main/napari/utils/_magicgui.py)
by registering a dropdown menu whenever a function parameter is annotated as one
of the basic napari [`Layer` types](https://napari.org/howtos/layers/index.html),
or, in this case, `ImageData` indicates we just the `data` attribute of the layer.
Furthermore, it recognizes when a function has a `napari.layers.Layer`
or `LayerData` return type annotation, and will add the result to the viewer.
So we gain a *lot* by annotating the above function with the appropriate
`napari` types.

```python
from napari.types import ImageData

def image_arithmetic(
    layerA: ImageData, operation: Operation, layerB: ImageData
) -> ImageData:
    return operation.value(layerA, layerB)
```

### the magic part

Finally, we decorate the function with `@magicgui` and tell it we'd like to
have a `call_button` that we can click to execute the function.

```python hl_lines="1"
@magicgui(call_button="execute")
def image_arithmetic(layerA: ImageData, operation: Operation, layerB: ImageData):
    return operation.value(layerA, layerB)
```

That's it!  The `image_arithmetic` function is now a
[FunctionGui][magicgui.widgets.FunctionGui] that can be shown, or incorporated
into other GUIs (such as the napari GUI shown in this example)

!!! note While [type hints](https://docs.python.org/3/library/typing.html)
    aren't always required in `magicgui`, they are recommended ... and they
    *are* required for certain things, like the `Operation(Enum)` [used here for
    the dropdown](#create-dropdowns-with-enums) and the `napari.types.ImageData`
    annotations that `napari` has registered with `magicgui`.

### create dropdowns with Enums

We'd like the user to be able to select the operation (`add`, `subtract`,
`multiply`, `divide`) using a dropdown menu. [`enum.Enum`][enum.Enum] offers a convenient
way to restrict values to a strict set of options, while providing `name: value`
pairs for each of the options. Here, the value for each choice is the actual
function we would like to have called when that option is selected.

```python
class Operation(enum.Enum):
    add = numpy.add
    subtract = numpy.subtract
    multiply = numpy.multiply
    divide = numpy.divide
```

### add it to napari

When we decorated the `image_arithmetic` function above, it became a
[FunctionGui][magicgui.widgets.FunctionGui].  Napari recognizes this type, so we
can simply add it to the napari viewer as follows:

```python
viewer.window.add_dock_widget(image_arithmetic)
```

### connect event listeners for interactivity

What fun is a GUI without some interactivity?  Let's make stuff happen.

We connect the `image_arithmetic.reset_choices` function to the
`viewer.layers.events.inserted/removed` event from `napari`, to make sure that
the dropdown menus stay in sync if a layer gets added or removed from the napari
window:

```python
viewer.layers.events.inserted.connect(image_arithmetic.reset_choices)
viewer.layers.events.removed.connect(image_arithmetic.reset_choices)
```

!!!tip
    An additional offering from `magicgui` here is that the decorated
    function also acquires a new attribute "`called`" that can be connected to
    callback functions of your choice.  Then, whenever the gui widget *or the
    original function* are called, the result will be passed to your callback
    function:

```python
@image_arithmetic.called.connect
def print_mean(value):
    # Callback function that accepts an event
    # the value attribute has the result of calling the function
    print(np.mean(value))
```

```
>>> image_arithmetic()
1.0060037881040373
```

## Code

Here's the full code example again.

"""

# %%
from enum import Enum

import napari
import numpy
from napari.types import ImageData

from magicgui import magicgui


class Operation(Enum):
    # A set of valid arithmetic operations for image_arithmetic.
    #
    # To create nice dropdown menus with magicgui, it's best
    # (but not required) to use Enums.  Here we make an Enum
    # class for all of the image math operations we want to
    # allow.
    add = numpy.add
    subtract = numpy.subtract
    multiply = numpy.multiply
    divide = numpy.divide


# here's the magicgui!  We also use the additional
# `call_button` option
@magicgui(call_button="execute")
def image_arithmetic(
    layerA: ImageData, operation: Operation, layerB: ImageData
) -> ImageData:
    # Add, subtracts, multiplies, or divides to image layers.
    return operation.value(layerA, layerB)


# create a viewer and add a couple image layers
viewer = napari.Viewer()
viewer.add_image(numpy.random.rand(20, 20), name="Layer 1")
viewer.add_image(numpy.random.rand(20, 20), name="Layer 2")

# add our new magicgui widget to the viewer
viewer.window.add_dock_widget(image_arithmetic)

# keep the dropdown menus in the gui in sync with the layer model
viewer.layers.events.inserted.connect(image_arithmetic.reset_choices)
viewer.layers.events.removed.connect(image_arithmetic.reset_choices)

napari.run()



import os
import sys
import time
from unittest.mock import patch

import pytest

from magicgui._util import debounce, user_cache_dir
from magicgui.widgets import FunctionGui


def test_user_cache_dir():
    ucd = user_cache_dir()
    import sys
    from pathlib import Path

    home = Path.home().resolve()
    if sys.platform == "win32":
        assert str(ucd) == str(home / "AppData" / "Local" / "magicgui" / "Cache")
    elif sys.platform == "darwin":
        assert str(ucd) == str(home / "Library" / "Caches" / "magicgui")
    else:
        assert str(ucd) == str(home / ".cache" / "magicgui")


@pytest.mark.skipif(
    bool(os.getenv("CI") and sys.platform == "win32"),
    reason="persistence test failing on CI",
)
def test_persistence(tmp_path):
    """Test that we can persist values across instances."""

    def _my_func(x: int, y="hello"): ...

    with patch("magicgui._util.user_cache_dir", lambda: tmp_path):
        fg = FunctionGui(_my_func, persist=True)
        assert str(tmp_path) in str(fg._dump_path)
        assert fg.x.value == 0
        fg.x.value = 10
        time.sleep(0.3)  # wait for debounce
        fg.y.value = "world"
        time.sleep(0.3)  # wait for debounce

        # second instance should match values of first
        fg2 = FunctionGui(_my_func, persist=True)
        assert fg2.x.value == 10
        assert fg2.y.value == "world"
        assert fg2.__signature__ == fg.__signature__
        assert fg2 is not fg


@pytest.mark.skipif(bool(os.getenv("CI")), reason="debounce test too brittle on CI")
def test_debounce():
    store = []

    @debounce(wait=0.1)
    def func(x):
        store.append(x)

    for i in range(10):
        func(i)
        time.sleep(0.034)
    time.sleep(0.15)

    assert len(store) <= 7  # exact timing will vary on CI ... fails too much
    assert store[-1] == 9



"""A wrapper around the tqdm.tqdm iterator that adds a ProgressBar to a magicgui."""

from __future__ import annotations

import contextlib
import inspect
from typing import Any, Iterable, cast

from magicgui.application import use_app
from magicgui.widgets import FunctionGui, ProgressBar

try:
    from tqdm import tqdm as _tqdm_std
except ImportError as e:  # pragma: no cover
    msg = (
        f"{e}. To use magicgui with tqdm please `pip install tqdm`, "
        "or use the tqdm extra: `pip install magicgui[tqdm]`"
    )
    raise type(e)(msg) from e


def _find_calling_function_gui(max_depth: int = 6) -> FunctionGui | None:
    """Traverse calling stack looking for a magicgui FunctionGui."""
    for finfo in inspect.stack()[2:max_depth]:
        # regardless of whether the function was decorated directly in the global module
        # namespace, or if it was renamed on decoration (`new_name = magicgui(func)`),
        # or if it was decorated inside of some local function scope...
        # we will eventually hit the :meth:`FunctionGui.__call__` method, which will
        # have the ``FunctionGui`` instance as ``self`` in its locals namespace.
        if finfo.function == "__call__" and finfo.filename.endswith("function_gui.py"):
            obj = finfo.frame.f_locals.get("self")
            return obj if isinstance(obj, FunctionGui) else None
    return None


_tqdm_kwargs = {
    p.name
    for p in inspect.signature(_tqdm_std.__init__).parameters.values()
    if p.kind is not inspect.Parameter.VAR_KEYWORD and p.name != "self"
}


class tqdm(_tqdm_std):
    """magicgui version of tqdm.

    See tqdm.tqdm API for valid args and kwargs: https://tqdm.github.io/docs/tqdm/

    Also, any keyword arguments to the
    [``magicgui.widgets.ProgressBar``][magicgui.widgets.ProgressBar] widget
    are also accepted and will be passed to the ``ProgressBar``.

    Examples
    --------
    When used inside of a magicgui-decorated function, ``tqdm`` (and the
    ``trange`` shortcut function) will append a visible progress bar to the gui
    container.

    >>> @magicgui(call_button=True)
    ... def long_running(steps=10, delay=0.1):
    ...     for i in tqdm(range(steps)):
    ...         sleep(delay)

    nesting is also possible:

    >>> @magicgui(call_button=True)
    ... def long_running(steps=10, repeats=4, delay=0.1):
    ...     for r in trange(repeats):
    ...         for s in trange(steps):
    ...             sleep(delay)
    """

    disable: bool

    def __init__(
        self, iterable: Iterable | None = None, *args: Any, **kwargs: Any
    ) -> None:
        kwargs = kwargs.copy()
        pbar_kwargs = {k: kwargs.pop(k) for k in set(kwargs) - _tqdm_kwargs}
        self._mgui = _find_calling_function_gui()
        if self._in_visible_gui:
            kwargs["gui"] = True
            kwargs.setdefault("mininterval", 0.025)
        super().__init__(iterable, *args, **kwargs)
        if not self._in_visible_gui:
            return

        # no-op status printer, required for older tqdm compat
        self.sp = lambda x: None
        if self.disable:
            return

        # check if we're being instantiated inside of a magicgui container
        self.progressbar = self._get_progressbar(**pbar_kwargs)
        self._app = use_app()

        if self.total is not None:
            # initialize progress bar range
            self.progressbar.range = (self.n, self.total)
            self.progressbar.value = self.n
        else:
            # show a busy indicator instead of a percentage of steps
            self.progressbar.range = (0, 0)
        self.progressbar.show()

    @property
    def _in_visible_gui(self) -> bool:
        try:
            return self._mgui is not None and self._mgui.visible
        except RuntimeError:
            return False

    def _get_progressbar(self, **kwargs: Any) -> ProgressBar:
        """Create ProgressBar or get from the parent gui `_tqdm_pbars` deque.

        The deque allows us to create nested iterables inside of a magigui, while
        resetting and reusing progress bars across ``FunctionGui`` calls. The nesting
        depth (into the deque) is reset by :meth:`FunctionGui.__call__`, right before
        the function is called.  Then, as the function encounters `tqdm` instances,
        this method gets or creates a progress bar and increment the
        :attr:`FunctionGui._tqdm_depth` counter on the ``FunctionGui``.
        """
        if self._mgui is None:
            return ProgressBar(**kwargs)

        if len(self._mgui._tqdm_pbars) > self._mgui._tqdm_depth:
            pbar = self._mgui._tqdm_pbars[self._mgui._tqdm_depth]
        else:
            pbar = ProgressBar(**kwargs)
            self._mgui._tqdm_pbars.append(pbar)
            self._mgui.append(pbar)
        self._mgui._tqdm_depth += 1
        return pbar

    def display(self, msg: str | None = None, pos: int | None = None) -> None:
        """Update the display."""
        if not self._in_visible_gui:
            super().display(msg=msg, pos=pos)
            return

        self.progressbar.value = self.n
        self._app.process_events()

    def close(self) -> None:
        """Cleanup and (if leave=False) close the progressbar."""
        if not self._in_visible_gui:
            super().close()
            return
        self._mgui = cast(FunctionGui, self._mgui)

        if self.disable:
            return

        # Prevent multiple closures
        self.disable = True

        # remove from tqdm instance set
        with self.get_lock():
            with contextlib.suppress(KeyError):
                self._instances.remove(self)
            if not self.leave:
                self._app.process_events()
                self.progressbar.hide()

        self._mgui._tqdm_depth -= 1


def trange(*args: Any, **kwargs: Any) -> tqdm:
    """Shortcut for tqdm(range(*args), **kwargs)."""
    return tqdm(range(*args), **kwargs)



"""# Callable functions demo

This example demonstrates handling callable functions with magicgui.
"""

from magicgui import magicgui


def f(x: int, y="a string") -> str:
    """Example function F."""
    return f"{y} {x}"


def g(x: int = 6, y="another string") -> str:
    """Example function G."""
    return f"{y} asdfsdf {x}"


@magicgui(call_button=True, func={"choices": ["f", "g"]})
def example(func="f"):
    """Ëxample function."""
    pass


def update(f: str):
    """Update function."""
    if len(example) > 2:
        del example[1]
    example.insert(1, magicgui(globals()[f]))


example.func.changed.connect(update)
example.show(run=True)



"""# File dialog widget

A dialog to select a file.
"""

from pathlib import Path

from magicgui import magicgui


@magicgui(filename={"mode": "r"})
def filepicker(filename=Path("~")) -> Path:
    """Take a filename and do something with it."""
    print("The filename is:", filename)
    return filename



filepicker.filename.changed.connect(print)
filepicker.show(run=True)



from magicgui.widgets.protocols import BaseApplicationBackend


class ApplicationBackend(BaseApplicationBackend):
    def _mgui_get_backend_name(self):
        return "ipynb"

    def _mgui_process_events(self):
        raise NotImplementedError()

    def _mgui_run(self):
        pass  # We run in IPython, so we don't run!

    def _mgui_quit(self):
        pass  # We don't run so we don't quit!

    def _mgui_get_native_app(self):
        return self

    def _mgui_start_timer(self, interval=0, on_timeout=None, single=False):
        raise NotImplementedError()

    def _mgui_stop_timer(self):
        raise NotImplementedError()



"""# Directory dialog widget

A dialog to select a directory.
"""

from pathlib import Path

from magicgui import magicgui


# Select a directory, instead of file(s)
@magicgui(directory={"mode": "d", "label": "Choose a directory"})
def directorypicker(directory=Path("~")):
    """Take a directory name and do something with it."""
    print("The directory name is:", directory)
    return directory


directorypicker.directory.changed.connect(print)
directorypicker.show(run=True)



from .application import ApplicationBackend
from .widgets import (
    CheckBox,
    ComboBox,
    Container,
    DateEdit,
    DateTimeEdit,
    EmptyWidget,
    FloatSlider,
    FloatSpinBox,
    Label,
    LineEdit,
    LiteralEvalLineEdit,
    Password,
    PushButton,
    RadioButton,
    Select,
    Slider,
    SpinBox,
    TextEdit,
    TimeEdit,
    ToolBar,
    get_text_width,
)

# DateTimeEdit
# show_file_dialog,

__all__ = [
    "ApplicationBackend",
    "CheckBox",
    "ComboBox",
    "Container",
    "DateEdit",
    "TimeEdit",
    "DateTimeEdit",
    "EmptyWidget",
    "FloatSlider",
    "FloatSpinBox",
    "Label",
    "LineEdit",
    "LiteralEvalLineEdit",
    "Password",
    "PushButton",
    "RadioButton",
    "Select",
    "Slider",
    "SpinBox",
    "TextEdit",
    "ToolBar",
    "get_text_width",
    "show_file_dialog",
]



import re
import subprocess
from itertools import count
from pathlib import Path
from textwrap import dedent

import mkdocs_gen_files
from qtpy.QtWidgets import QApplication

DOCS = Path(__file__).parent.parent
CODE_BLOCK = re.compile("``` ?python\n([^`]*)```", re.DOTALL)


def _set_dark_mode(dark_mode: bool):
    """Set the system to dark mode or not."""
    cmd = ["osascript", "-l", "JavaScript", "-e"]
    if dark_mode:
        cmd += ["Application('System Events').appearancePreferences.darkMode = true"]
    else:
        cmd += ["Application('System Events').appearancePreferences.darkMode = false"]
    subprocess.run(cmd)


# internally used trick to leave a window open at the end of a cell
# add this to the end of a cell to leave it open
LEAVE_OPEN = "# leave open"
LEFT_OPEN = False


def _clear_left_open() -> None:
    global LEFT_OPEN
    for w in QApplication.topLevelWidgets():
        w.close()
        w.deleteLater()
    QApplication.processEvents()
    LEFT_OPEN = False


def _write_markdown_result_image(src: str, ns: dict, dest: str) -> None:
    global LEFT_OPEN
    top_widgets = set(QApplication.topLevelWidgets())
    try:
        exec(src, ns, ns)
    except NameError as e:
        raise NameError(
            f"Error evaluating code for {dest!r} with namespace {set(ns)!r}:\n\n{src}. "
            f"{e}"
        ) from e

    if not LEFT_OPEN:
        top_widgets = set(QApplication.topLevelWidgets()) - top_widgets
    if not top_widgets:
        print("No top level widgets found for", dest)
    if len(top_widgets) > 1:
        name = re.search(r"([^\s\.]+).show\(\)", src, re.DOTALL)[1]
        try:
            top_widgets = {ns[name].native}
        except KeyError:
            pass  # FIXME... it's probably an attribute
    for w in top_widgets:
        w.setMinimumWidth(200)
        w.show()
        w.activateWindow()
        w.update()

        with mkdocs_gen_files.open(dest, "wb") as f:
            if not w.grab().save(f.name):
                print("Error saving", dest)

        if LEAVE_OPEN in src:
            LEFT_OPEN = True
        else:
            _clear_left_open()


EXCLUDE = {"migration.md"}


def main() -> None:
    for mdfile in sorted(DOCS.rglob("*.md"), reverse=True):
        if mdfile.name in EXCLUDE:
            continue
        _clear_left_open()

        md = mdfile.read_text()
        w_iter = count()
        code_blocks = list(CODE_BLOCK.finditer(md))
        has_show = any(".show()" in match.group(0) for match in code_blocks)
        if has_show:
            namespace: dict = {}
            for match in code_blocks:
                code = dedent(match.group(1))
                if ".show()" in match.group(0):
                    dest = f"_images/{mdfile.stem}_{next(w_iter)}.png"
                    _write_markdown_result_image(code, namespace, dest)
                else:
                    # still need to execute the code to populate the namespace
                    # for later code blocks
                    exec(code, namespace, namespace)


main()
